#emacs, this is -*-Python-*- mode
#cython: language_level=2

import numpy as np
cimport numpy as np
cimport _fastgeom
cimport _mahalanobis
cimport _pmat_jacobian

import numpy
import numpy as np
import time, sys
import adskalman.adskalman as kalman

import flydra_core.kalman.ekf as kalman_ekf
#import flydra_core.geom as geom
import _fastgeom as geom
import flydra_core.geom
import _mahalanobis
import math, struct
import flydra_core.data_descriptions
from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood
import collections
from pprint import pprint

cdef double c_inf
c_inf = np.inf

__all__ = ['TrackedObject']

PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3
PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4

PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX

NO_LCOORDS = numpy.nan,numpy.nan,numpy.nan,  numpy.nan,numpy.nan,numpy.nan

cdef double cnan
cnan = np.nan

ctypedef int mybool

cdef extern from "math.h":
    double sqrt(double)

cpdef evaluate_pmat_jacobian(object pmats_and_points_cov, np.ndarray[np.double_t, ndim=1] xhatminus):
    cdef int N
    cdef mybool missing_data
    cdef int i
    cdef int ss
    cdef _pmat_jacobian.PinholeCameraModelWithJacobian pinhole_model
    cdef np.ndarray[np.double_t, ndim=1] y
    cdef np.ndarray[np.double_t, ndim=2] C
    cdef np.ndarray[np.double_t, ndim=1] hx
    cdef np.ndarray[np.double_t, ndim=2] R
    cdef np.ndarray[np.double_t, ndim=1] hx_i

    ss = xhatminus.shape[0]

    N = len(pmats_and_points_cov) # number of observations
    if N > 0:
        missing_data = False
    else:
        missing_data = True

    # Create 2N vector of N observations.
    y = numpy.empty((2*N,), dtype=numpy.float64)

    # Create 2N x 4 observation model matrix (jacobian of h() at xhatminus).
    C = numpy.zeros((2*N,ss), dtype=numpy.float64)

    # Create 2N vector h(xhatminus) where xhatminus is the a
    # priori estimate and h() is the observation model.
    hx = numpy.empty((2*N,), dtype=numpy.float64)

    # Create 2N x 2N observation covariance matrix
    R = numpy.zeros((2*N,2*N), dtype=numpy.float64)

    # evaluate jacobian for each participating camera
    for i,(nonlin_model,xy2d_obs,cov) in enumerate(pmats_and_points_cov):
        pinhole_model = nonlin_model

        # fill prediction vector [ h(xhatminus) ]
        hx_i = pinhole_model.evaluate(xhatminus[:3])
        hx[2*i:2*i+2] = hx_i

        # fill observation vector
        y[2*i:2*i+2] = xy2d_obs

        # fill observation model
        C_i = pinhole_model.evaluate_jacobian_at(xhatminus[:3])
        C[2*i:2*i+2,:3] = C_i

        # fill observation covariance
        R[2*i:2*i+2,2*i:2*i+2]=cov

    ## if 1:
    ##     print 'C'
    ##     print C
    ##     print 'y'
    ##     print y
    ##     print 'hx'
    ##     print hx
    ##     print 'R'
    ##     print R
    return y, hx, C, R, missing_data

def obs2d_hashable( arr ):
    # XXX FIXME: this could likely be sped up
    assert arr.dtype == numpy.uint16
    assert len(arr.shape)==1

    # sort array based on camn
    camns = arr[0::2]
    pt_idx = arr[1::2]
    camn_order = numpy.argsort(camns)

    # fill new array with sorted values
    newarr = numpy.zeros( (len(arr),), dtype = numpy.uint16 )
    newarr[ camn_order*2 ] = camns
    newarr[ camn_order*2+1 ] = pt_idx

    val = newarr.tostring()
    return val

cdef class TrackedObject:
    """
    Track one object using a Kalman filter.

    """
    cdef readonly long current_frameno
    cdef readonly long last_frameno_with_data
    cdef long long max_frames_skipped
    cdef readonly mybool kill_me
    cdef mybool save_all_data
    cdef double area_threshold, area_threshold_for_orientation

    cdef object reconstructor, my_kalman
    cdef object distorted_pixel_euclidian_distance_accept
    cdef double max_variance
    cdef object ekf_observation_covariance_pixels

    cdef readonly object frames, xhats, timestamps, Ps, MLE_position, MLE_Lcoords
    cdef readonly object observations_frames, observations_2d
    cdef int disable_image_stat_gating, orientation_consensus
    cdef object fake_timestamp
    cdef readonly unsigned int obj_id
    cdef object ekf_kalman_A, ekf_kalman_Q

    def __init__(self,
                 reconstructor, # the Reconstructor instance
                 obj_id,
                 long frame, # frame number of first data
                 obs0_position, # first data
                 obs0_Lcoords, # first data
                 first_observation_camns,
                 first_observation_idxs,
                 kalman_model=None,
                 save_all_data=False,
                 double area_threshold=0.0,
                 double area_threshold_for_orientation=0.0,
                 disable_image_stat_gating=False,
                 orientation_consensus=0,
                 fake_timestamp=None
                 ):
        """

        arguments
        =========
        reconstructor - reconstructor instance
        obj_id - unique identifier for each object
        frame - frame number of first observation data
        obs0_position - first observation (in arbitrary units)
        kalman_model - Kalman parameters
        area_threshold - minimum area to consider for tracking use
        """
        self.obj_id = obj_id
        self.area_threshold = area_threshold
        self.area_threshold_for_orientation = area_threshold_for_orientation
        self.save_all_data = save_all_data
        self.kill_me = False
        self.reconstructor = reconstructor
        self.distorted_pixel_euclidian_distance_accept=kalman_model.get('distorted_pixel_euclidian_distance_accept',None)
        self.disable_image_stat_gating = disable_image_stat_gating
        self.orientation_consensus = orientation_consensus
        self.fake_timestamp = fake_timestamp

        self.current_frameno = frame
        self.last_frameno_with_data = frame
        obs0_position = obs0_position
        if obs0_Lcoords is None:
            obs0_Lcoords = NO_LCOORDS
        else:
            PL_line3d = flydra_core.geom.line_from_HZline(obs0_Lcoords) # PlueckerLine instance
            loc = PL_line3d.closest() # closest point on line to origin
            line3d = flydra_core.geom.line_from_points( loc, loc+PL_line3d.direction() )
            obs0_Lcoords = line3d.to_hz()
        ss = kalman_model['ss']
        initial_x = numpy.zeros( (ss,) )
        initial_x[:3] = obs0_position
        P_k1=numpy.eye(ss) # initial state error covariance guess
        for i in range(0,3):
            P_k1[i,i]=kalman_model['initial_position_covariance_estimate']
        for i in range(3,6):
            P_k1[i,i]=kalman_model.get('initial_velocity_covariance_estimate',0)
        if ss >= 9:
            for i in range(6,9):
                P_k1[i,i]=kalman_model['initial_acceleration_covariance_estimate']

        self.max_variance = kalman_model['max_variance_dist_meters']**2 # square so that it is in variance units

        self.max_frames_skipped = kalman_model['max_frames_skipped']

        if kalman_model.get( 'isEKF', False):
            self.ekf_kalman_A = kalman_model['A']
            self.ekf_kalman_Q = kalman_model['Q']
            # EKF
            self.my_kalman = kalman_ekf.EKF(
                initial_x=initial_x,
                initial_P=P_k1,
                )
            self.ekf_observation_covariance_pixels = kalman_model['ekf_observation_covariance_pixels']
        else:
            # non-EKF
            self.my_kalman = kalman.KalmanFilter(kalman_model['A'],
                                                 kalman_model['C'],
                                                 kalman_model['Q'],
                                                 kalman_model['R'],
                                                 initial_x,
                                                 P_k1)
        self.frames = [frame]
        self.xhats = [initial_x]
        if self.fake_timestamp is None:
            self.timestamps = [time.time()]
        else:
            self.timestamps = [self.fake_timestamp]
        self.Ps = [P_k1]

        self.observations_frames = [frame]
        self.MLE_position = [obs0_position]
        self.MLE_Lcoords = [obs0_Lcoords]

        first_observations_2d_pre = [[camn,idx] for camn,idx in zip(first_observation_camns,first_observation_idxs)]
        first_observations_2d = []
        for obs in first_observations_2d_pre:
            first_observations_2d.extend( obs )
        first_observations_2d = numpy.array(first_observations_2d,dtype=numpy.uint16) # if saved as VLArray, should match with atom type

        self.observations_2d = [first_observations_2d]

        self.max_frames_skipped=kalman_model['max_frames_skipped']

        # Don't run kalman filter with initial data, as this would
        # cause error estimates to drop too low.

    def __repr__(self):
        return '<TRO frames[0]=%r observations_2d[0]=%r xhats[0]=%r>' % (
            self.frames[0], self.observations_2d[0], self.xhats[0])

    def debug_info(self,level=3):
        if level > 5:
            sys.stdout.write('%s\n'%self)
            N_pts = len(self.xhats)
            start_idx = max( N_pts-10, 0 )
            for i in range(start_idx,N_pts):
                this_Pmean = math.sqrt(self.Ps[i][0,0]**2 + self.Ps[i][1,1]**2 + self.Ps[i][2,2]**2)
                sys.stdout.write( ' '.join(map(str,['  ',i,self.frames[i],self.xhats[i][:3],this_Pmean,])) )
                if self.frames[i] in self.observations_frames:
                    j =  self.observations_frames.index(  self.frames[i] )
                    sys.stdout.write( '%s\n'%self.MLE_position[j] )
                else:
                    sys.stdout.write('\n')
            sys.stdout.write('\n')
        elif level > 2:
            sys.stdout.write('%d observations, %d estimates for %s\n'%(len(self.xhats),len(self.MLE_position),self))

    def get_distance_and_nsigma( self, testx ):
        xhat = self.xhats[-1][:3]

        dist2 = numpy.sum((testx-xhat)**2) # distance squared
        dist = numpy.sqrt(dist2)

        # XXX This should be mahalanobis distance, probably.
        P = self.Ps[-1]
        Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)])) # sigma squared
        sigma = numpy.sqrt(Pmean)
        return dist, (dist/sigma)

    def kill(self):
        # called when killed
        if self.save_all_data:
            return

        # find last data
        last_observation_frame = self.observations_frames[-1]

        # eliminate estimates past last observation
        while 1:
            if self.frames[-1] > last_observation_frame:
                self.frames.pop()
                self.xhats.pop()
                self.timestamps.pop()
                self.Ps.pop()
            else:
                break

    cpdef calculate_a_posteriori_estimate(self,
                                        long frame,
                                        object data_dict,
                                        object camn2cam_id,
                                        int debug1=0,
                                        int skip_data_association=0,
                                        object original_camns_and_idxs=None,
                                        object original_cam_ids_and_points2d=None,
                                        ):
        # Step 1. Update Kalman state to a priori estimates for this frame.
        # Step 1.A. Update Kalman state for each skipped frame.

        # Make sure we have xhat_k1 (previous frames' a posteriori)

        # For each frame that was skipped, step the Kalman filter.
        # Since we have no observation, the estimated error will
        # rise.
        cdef long i, frames_since_update
        cdef long frames_skipped
        cdef double Pmean
        frames_since_update = frame-self.current_frameno-1

        if debug1>2:
            print 'doing',self,'============--'
            print 'updating for %d frames since update'%(frames_since_update,)

        for i in range(frames_since_update):
            if isinstance(self.my_kalman, kalman_ekf.EKF):
                xhat, P = self.my_kalman.step(self.ekf_kalman_A,
                                              self.ekf_kalman_Q)
            else:
                xhat, P = self.my_kalman.step()
            ############ save outputs ###############
            self.frames.append( self.current_frameno + i + 1 )
            self.xhats.append( xhat )
            self.timestamps.append( 0.0 )
            self.Ps.append( P )

        self.current_frameno = frame
        this_observations_2d_hash = None
        used_camns_and_idxs = []
        all_close_camn_pt_idxs = []
        Pmean = c_inf
        if not self.kill_me:
            # Step 1.B. Update Kalman to provide a priori estimates for this frame
            if isinstance(self.my_kalman, kalman_ekf.EKF):
                xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori(
                    self.ekf_kalman_A, self.ekf_kalman_Q)
            else:
                xhatminus, Pminus = self.my_kalman.step1__calculate_a_priori()
            if debug1>2:
                print 'xhatminus'
                print xhatminus
                print 'Pminus'
                print Pminus
                print

            # Step 2. Filter incoming 2D data to use informative points (data association)
            if skip_data_association:
                position_MLE = numpy.nan*numpy.ones( (3,)) # skip
                Lcoords = None
                used_camns_and_idxs = original_camns_and_idxs
                cam_ids_and_points2d = original_cam_ids_and_points2d
                all_close_camn_pt_idxs = [] # not important when skipping data association
            else:
                (position_MLE, Lcoords, used_camns_and_idxs,
                 cam_ids_and_points2d,
                 all_close_camn_pt_idxs) = self._filter_data(xhatminus, Pminus,
                                                             data_dict,
                                                             camn2cam_id,
                                                             debug=debug1)
            if debug1>2:
                print 'position MLE, used_camns_and_idxs',position_MLE,used_camns_and_idxs
                print 'Lcoords (3D body orientation) : %s'%str(Lcoords)

            # Step 3. Incorporate observation to estimate a posteriori
            if isinstance(self.my_kalman, kalman_ekf.EKF):
                prediction_3d = xhatminus[:3]
                pmats_and_points_cov = [ (
                                          self.reconstructor.get_model_with_jacobian(cam_id),
                                          value_tuple[:2],#just first 2 components (x,y) become xy2d_observed
                                          self.ekf_observation_covariance_pixels)
                                         for (cam_id,value_tuple) in cam_ids_and_points2d]
                y,hx,C,R,missing_data = evaluate_pmat_jacobian(
                    pmats_and_points_cov,xhatminus)
                xhat, P = self.my_kalman.step2__calculate_a_posteriori(
                    xhatminus, Pminus, y=y,hx=hx,
                    C=C,R=R,missing_data=missing_data)
            else:
                xhat, P = self.my_kalman.step2__calculate_a_posteriori(
                    xhatminus, Pminus,
                    y=position_MLE)

            # calculate mean variance of x y z position (assumes first three components of state vector are position)

            # XXX should probably use trace/N (mean of variances) or determinant (volume of variance)
            Pmean = numpy.sqrt(numpy.sum([P[i,i]**2 for i in range(3)]))

            if debug1>2:
                print 'xhat'
                print xhat
                print 'P'
                print P
                print 'Pmean',Pmean

            # XXX Need to test if error for this object has grown beyond a
            # threshold at which it should be terminated.
            if Pmean > self.max_variance:
                self.kill_me = True
                if debug1>=1:
                    print 'will kill next time because Pmean too large (%f > %f)'%(Pmean,self.max_variance)

            frames_skipped = frame - self.last_frameno_with_data
            if frames_skipped > self.max_frames_skipped:
                self.kill_me = True
                if debug1>=1:
                    print 'will kill next time because frames skipped (%ld > %ld, frame %ld)'%(
                        frames_skipped, self.max_frames_skipped,frame)

            ############ save outputs ###############
            self.frames.append( frame )
            self.xhats.append( xhat )
            if self.fake_timestamp is None:
                self.timestamps.append(time.time())
            else:
                self.timestamps.append(self.fake_timestamp)
            self.Ps.append( P )

            if position_MLE is not None:
                self.last_frameno_with_data = frame
                self.observations_frames.append( frame )
                self.MLE_position.append( position_MLE )
                if Lcoords is None:
                    self.MLE_Lcoords.append( NO_LCOORDS )
                else:
                    self.MLE_Lcoords.append( Lcoords )
                this_observations_2d = []
                for (camn, frame_pt_idx, dd_idx) in used_camns_and_idxs:
                    this_observations_2d.extend( [camn,frame_pt_idx] )
                this_observations_2d = numpy.array( this_observations_2d, dtype=numpy.uint16 ) # convert to numpy
                self.observations_2d.append( this_observations_2d )
                this_observations_2d_hash = obs2d_hashable( this_observations_2d )
            if debug1>2:
                print
        return (used_camns_and_idxs, this_observations_2d_hash,
                Pmean, all_close_camn_pt_idxs)

    def remove_previous_observation(self, int debug1=0):

        # This will remove the just-done observation from my
        # state. Thus, this instance will act as it it skipped data on
        # one observation.

        self.current_frameno -= 1

        # Remove most recent information
        frame = self.frames.pop()
        if debug1>2:
            sys.stdout.write( '%s removing previous observation (from frame %d)\n'%(self,frame,))

        self.xhats.pop()
        self.timestamps.pop()
        self.Ps.pop()
        if self.observations_frames[-1] == frame:
            self.observations_frames.pop()
            self.MLE_position.pop()
            self.MLE_Lcoords.pop()
            self.observations_2d.pop()

        # reset self.my_kalman
        self.my_kalman.xhat_k1 = self.xhats[-1]
        self.my_kalman.P_k1 = self.Ps[-1]

        self.last_frameno_with_data = self.observations_frames[-1]

    cpdef _filter_data(self, object xhatminus, object Pminus,
                       object data_dict, object camn2cam_id,
                       int debug=0):
        """given state estimate, select useful incoming data and make new observation

        This function 'solves' the data association problem. 2D
        observations are associated with a kalman object if they are
        with some distance between a predicted image of the object on
        the camera plane. The distance estimation is made in 3D as the
        Euclidian distance between the ray in 3D space corresponding
        to the image point observed and the predicted 3D location of
        the kalman object. (Other distance metrics could also be
        implemented, such as 2D Euclidian distance on the image
        plane. Note that a raw threshold for this is implemented with
        distorted_pixel_euclidian_distance_accept.)

        """
        # For each camera, predict 2D image location and error distance
        cdef double least_nll, nll_this_point
        cdef double dist2, dist, p_y_x
        cdef int gated_in, pixel_dist_criterion_passed

        cdef double pt_area, mean_val, sumsqf_val, area
        cdef int cur_val
        cdef int camn, frame_pt_idx
        cdef _fastgeom.PlueckerLine projected_line
        cdef _fastgeom.ThreeTuple best_3d_location

        all_close_camn_pt_idxs = [] # store all "maybes"

        prediction_3d = xhatminus[:3]
        cdef _fastgeom.ThreeTuple fast_prediction_3d = _fastgeom.ThreeTuple(xhatminus[:3])
        pixel_dist_cmp = self.distorted_pixel_euclidian_distance_accept
        neg_predicted_3d = -geom.ThreeTuple( prediction_3d )
        cam_ids_and_points2d = []

        used_camns_and_idxs = []
        if debug>2:
            print '_filter_data():'
        for camn, candidate_point_list in data_dict.items():
            cam_id = camn2cam_id[camn]

            if pixel_dist_cmp is not None:
                predicted_2d_distorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=True)

            if debug>2:
                predicted_2d_undistorted = self.reconstructor.find2d(cam_id,prediction_3d,distorted=False)
                print '  cam_id',cam_id,'camn',camn,'--------'
                print '    predicted_2d (undistorted)',predicted_2d_undistorted

            # Iterate over all 2D data. Find the 2D point which minimizes::
            #    f(area, maxabsdiff, etc) + d(y,xhat).

            # According to Kristin, this is equivalent to choosing the
            # point with maximum cumulative probability::
            #    p( y_t | x_t ) p( x_t, x_{t-1})

            # [In this case f(area,...) represents negative log of
            # the likelihood of observation y given x.]

            # For large numbers of 2d points in data_dict, probably
            # faster to compute 2d image of error ellipsoid and see if
            # data_dict points fall inside that. For now, however,
            # compute distance in 3d. This avoids having to do a
            # (nonlinear) perspective projection of a multivariante
            # normal.

            least_nll = c_inf
            closest_idx = None

            Pminus_inv = None # defer inverting Pminus until necessary.

            for idx,(pt_undistorted,projected_line) in enumerate(
                candidate_point_list):

                # Iterate over each candidate point. Each point has:
                #  * index 'idx'
                #  * a bunch of data (such as 2D coordinates x and y)
                #    'pt_undistorted',
                #  * and the 3D line passing through that ray and the
                #    image plane.

                gated_in = False

                # First, a quick gating based on image plane distance.

                pixel_dist_criterion_passed = True
                if pixel_dist_cmp is not None:
                    # XXX TODO: fixme: should just pass in distorted pixel coordinates, but saves reorganizing all this code.
                    pt_x_undist =  pt_undistorted[PT_TUPLE_IDX_X]
                    pt_y_undist =  pt_undistorted[PT_TUPLE_IDX_Y]
                    pt_x_dist, pt_y_dist = self.reconstructor.distort( cam_id, (pt_x_undist, pt_y_undist) )
                    pixel_dist = numpy.sqrt((predicted_2d_distorted[0] - pt_x_dist)**2 + (predicted_2d_distorted[1] - pt_y_dist)**2)
                    if pixel_dist > pixel_dist_cmp:
                        pixel_dist_criterion_passed = False

                if pixel_dist_criterion_passed:

                    # Second, a quick gating based on image attributes.

                    pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]
                    tmp = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
                    if tmp is None:
                        if not self.disable_image_stat_gating:
                            raise ValueError(
                                '--disable-image-stat-gating not specified and '
                                'no image statistics were saved.')
                    else:
                        cur_val = tmp
                        mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
                        sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]

                    if cur_val is None or self.disable_image_stat_gating:
                        p_y_x = 0.0 # no penalty
                    else:
                        # this could even depend on 3d geometry
                        p_y_x = some_rough_negative_log_likelihood(
                            pt_area, cur_val, mean_val, sumsqf_val )

                    if pt_area < self.area_threshold:
                        # This should not fall under
                        # "disable_image_stat_gating" -- it is a separate test.
                        p_y_x = np.inf

                    if np.isfinite(p_y_x):
                        gated_in = True

                        # Find point on ray with closest Mahalanobis distance.

                        if Pminus_inv is None:
                            Pminus_inv = numpy.linalg.inv( Pminus[:3,:3] )
                        best_3d_location = _mahalanobis.line_fit_3d(
                            projected_line, fast_prediction_3d, Pminus_inv )

                        # find closest distance between projected_line and predicted position for each 2d point
                        #   squared distance between prediction and camera ray
                        dist2=_mahalanobis.dist2( best_3d_location,
                                                  fast_prediction_3d,
                                                  Pminus_inv )
                        dist = sqrt(dist2)
                        nll_this_point = p_y_x + dist # negative log likelihood of this point

                frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
                if debug>2:
                    extra_print = []
                    if not self.disable_image_stat_gating:
                        cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
                        mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
                        sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
                        if pixel_dist_criterion_passed:
                            extra_print.append('(pt_area %f, cur %d, mean %.1f, sumsqf %.1f)'%(
                                pt_area,cur_val,mean_val,sumsqf_val))

                    if gated_in:
                        extra_print.append('distorted %.1f %.1f (pixel_dist = %.1f, mahal dist = %.1f, criterion passed=%s)'%(
                            pt_x_dist, pt_y_dist, pixel_dist, dist, str(pixel_dist_criterion_passed)))
                    print '    ->', dist2, pt_undistorted[:2], '(idx %d) %s'%(
                        frame_pt_idx,' '.join(extra_print))

                if gated_in:
                    if debug>2:
                        print '       (acceptable)'
                    if nll_this_point < least_nll:
                        if debug>2:
                            print '       (so far the best -- taking)'
                        least_nll = nll_this_point
                        closest_idx = idx
                    all_close_camn_pt_idxs.append( (camn, frame_pt_idx) )
                elif debug>2:
                    print '       (not acceptable)'

            if closest_idx is not None:
                pt_undistorted, projected_line = candidate_point_list[closest_idx]
                area = pt_undistorted[PT_TUPLE_IDX_AREA]
                if area >= self.area_threshold_for_orientation:
                    # with orientation
                    observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
                                   pt_undistorted[PT_TUPLE_IDX_Y],
                                   pt_undistorted[PT_TUPLE_IDX_AREA],
                                   pt_undistorted[PT_TUPLE_IDX_SLOPE],
                                   pt_undistorted[PT_TUPLE_IDX_ECCENTRICITY],
                                   pt_undistorted[PT_TUPLE_IDX_P1],
                                   pt_undistorted[PT_TUPLE_IDX_P2],
                                   pt_undistorted[PT_TUPLE_IDX_P3],
                                   pt_undistorted[PT_TUPLE_IDX_P4])
                else:
                    # with no orientation
                    observed_2d = (pt_undistorted[PT_TUPLE_IDX_X],
                                   pt_undistorted[PT_TUPLE_IDX_Y],
                                   pt_undistorted[PT_TUPLE_IDX_AREA],
                                   cnan,
                                   cnan,
                                   cnan,
                                   cnan,
                                   cnan,
                                   cnan)
                cam_ids_and_points2d.append( (cam_id,observed_2d) )
                frame_pt_idx = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
                used_camns_and_idxs.append( (camn, frame_pt_idx, closest_idx) )
                if debug>2:
                    print 'best match idx %d (%s)'%(closest_idx, str(pt_undistorted[:2]))

        Lcoords = None # default to no line coordinates
        # Now cam_ids_and_points2d has just the 2d points we'll use for this reconstruction
        if len(cam_ids_and_points2d)==1:
            # keep 3D "observation" because we need to save 2d observations
            position_MLE = numpy.nan*numpy.ones( (3,))
        elif len(cam_ids_and_points2d)>=2:
            if self.reconstructor.wateri is not None:
                # do not attempt this if we have refractive boundary
                # keep 3D "observation" because we need to save 2d observations
                position_MLE = numpy.nan*numpy.ones( (3,))
            else:
                position_MLE, Lcoords = self.reconstructor.find3d(
                    cam_ids_and_points2d, return_line_coords = True,
                    orientation_consensus=self.orientation_consensus)
        else:
            position_MLE = None
        return (position_MLE, Lcoords, used_camns_and_idxs,
                cam_ids_and_points2d, all_close_camn_pt_idxs)

    def get_most_recent_data(self):
        if not len(self.xhats):
            return
        xhat = self.xhats[-1]
        P = self.Ps[-1]
        return self.obj_id, xhat,P



import tables as PT
import numpy as np

import flydra_core.data_descriptions
import flydra_core.kalman.dynamic_models
from flydra_core.kalman.point_prob import some_rough_negative_log_likelihood

PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
PT_TUPLE_IDX_AREA = flydra_core.data_descriptions.PT_TUPLE_IDX_AREA
PT_TUPLE_IDX_SLOPE = flydra_core.data_descriptions.PT_TUPLE_IDX_SLOPE
PT_TUPLE_IDX_ECCENTRICITY = flydra_core.data_descriptions.PT_TUPLE_IDX_ECCENTRICITY
PT_TUPLE_IDX_P1 = flydra_core.data_descriptions.PT_TUPLE_IDX_P1
PT_TUPLE_IDX_P2 = flydra_core.data_descriptions.PT_TUPLE_IDX_P2
PT_TUPLE_IDX_P3 = flydra_core.data_descriptions.PT_TUPLE_IDX_P3
PT_TUPLE_IDX_P4 = flydra_core.data_descriptions.PT_TUPLE_IDX_P4

PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX


class KalmanEstimates(PT.IsDescription):
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.Float64Col(pos=2)  # time of reconstruction
    x = PT.Float32Col(pos=3)
    y = PT.Float32Col(pos=4)
    z = PT.Float32Col(pos=5)
    xvel = PT.Float32Col(pos=6)
    yvel = PT.Float32Col(pos=7)
    zvel = PT.Float32Col(pos=8)
    xaccel = PT.Float32Col(pos=9)
    yaccel = PT.Float32Col(pos=10)
    zaccel = PT.Float32Col(pos=11)
    # save diagonal of P matrix
    P00 = PT.Float32Col(pos=12)
    P11 = PT.Float32Col(pos=13)
    P22 = PT.Float32Col(pos=14)
    P33 = PT.Float32Col(pos=15)
    P44 = PT.Float32Col(pos=16)
    P55 = PT.Float32Col(pos=17)
    P66 = PT.Float32Col(pos=18)
    P77 = PT.Float32Col(pos=19)
    P88 = PT.Float32Col(pos=20)


class KalmanEstimatesVelOnly(PT.IsDescription):
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.Float64Col(pos=2)  # time of reconstruction
    x = PT.Float32Col(pos=3)
    y = PT.Float32Col(pos=4)
    z = PT.Float32Col(pos=5)
    xvel = PT.Float32Col(pos=6)
    yvel = PT.Float32Col(pos=7)
    zvel = PT.Float32Col(pos=8)
    # save diagonal of P matrix
    P00 = PT.Float32Col(pos=9)
    P11 = PT.Float32Col(pos=10)
    P22 = PT.Float32Col(pos=11)
    P33 = PT.Float32Col(pos=12)
    P44 = PT.Float32Col(pos=13)
    P55 = PT.Float32Col(pos=14)


class KalmanEstimatesVelOnlyWithDirection(PT.IsDescription):
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.Float64Col(pos=2)  # time of reconstruction
    x = PT.Float32Col(pos=3)
    y = PT.Float32Col(pos=4)
    z = PT.Float32Col(pos=5)
    xvel = PT.Float32Col(pos=6)
    yvel = PT.Float32Col(pos=7)
    zvel = PT.Float32Col(pos=8)
    # save diagonal of P matrix
    P00 = PT.Float32Col(pos=9)
    P11 = PT.Float32Col(pos=10)
    P22 = PT.Float32Col(pos=11)
    P33 = PT.Float32Col(pos=12)
    P44 = PT.Float32Col(pos=13)
    P55 = PT.Float32Col(pos=14)
    # save estimated direction of long body axis
    rawdir_x = PT.Float32Col(pos=15)
    rawdir_y = PT.Float32Col(pos=16)
    rawdir_z = PT.Float32Col(pos=17)
    dir_x = PT.Float32Col(pos=18)
    dir_y = PT.Float32Col(pos=19)
    dir_z = PT.Float32Col(pos=20)


class KalmanEstimatesVelOnlyPositionCovariance(PT.IsDescription):
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.Float64Col(pos=2)  # time of reconstruction
    x = PT.Float32Col(pos=3)
    y = PT.Float32Col(pos=4)
    z = PT.Float32Col(pos=5)
    xvel = PT.Float32Col(pos=6)
    yvel = PT.Float32Col(pos=7)
    zvel = PT.Float32Col(pos=8)
    # save parts of P matrix
    P00 = PT.Float32Col(pos=9)
    P01 = PT.Float32Col(pos=10)
    P02 = PT.Float32Col(pos=11)
    P11 = PT.Float32Col(pos=12)
    P12 = PT.Float32Col(pos=13)
    P22 = PT.Float32Col(pos=14)
    P33 = PT.Float32Col(pos=15)
    P44 = PT.Float32Col(pos=16)
    P55 = PT.Float32Col(pos=27)


class KalmanEstimatesVelOnlyWithDirectionPositionCovariance(PT.IsDescription):
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.Float64Col(pos=2)  # time of reconstruction
    x = PT.Float32Col(pos=3)
    y = PT.Float32Col(pos=4)
    z = PT.Float32Col(pos=5)
    xvel = PT.Float32Col(pos=6)
    yvel = PT.Float32Col(pos=7)
    zvel = PT.Float32Col(pos=8)
    # save parts of P matrix
    P00 = PT.Float32Col(pos=9)
    P01 = PT.Float32Col(pos=10)
    P02 = PT.Float32Col(pos=11)
    P11 = PT.Float32Col(pos=12)
    P12 = PT.Float32Col(pos=13)
    P22 = PT.Float32Col(pos=14)
    P33 = PT.Float32Col(pos=15)
    P44 = PT.Float32Col(pos=16)
    P55 = PT.Float32Col(pos=17)
    # save estimated direction of long body axis
    rawdir_x = PT.Float32Col(pos=18)
    rawdir_y = PT.Float32Col(pos=19)
    rawdir_z = PT.Float32Col(pos=20)
    dir_x = PT.Float32Col(pos=21)
    dir_y = PT.Float32Col(pos=22)
    dir_z = PT.Float32Col(pos=23)


class KalmanSaveInfo(object):
    def __init__(
        self,
        name=None,
        allocate_space_for_direction=False,
        save_covariance="position"  # or 'full' or 'diag'
        # 'position' means to save full covariance for upper 3x3, diag for rest
    ):
        if save_covariance == "full":
            raise NotImplementedError("")
        assert save_covariance in ["position", "diag"]
        self.save_covariance = save_covariance
        model = flydra_core.kalman.dynamic_models.get_kalman_model(name=name, dt=1.0)
        ss = model["ss"]
        if ss == 6:
            if self.save_covariance == "diag":
                if allocate_space_for_direction:
                    self.description = KalmanEstimatesVelOnlyWithDirection
                else:
                    self.description = KalmanEstimatesVelOnly
            else:
                # 'position'
                if allocate_space_for_direction:
                    self.description = (
                        KalmanEstimatesVelOnlyWithDirectionPositionCovariance
                    )
                else:
                    self.description = KalmanEstimatesVelOnlyPositionCovariance
                # indices for upper part of covariance matrix (which is symmetric, so this is sufficient)
                self.rows = np.array([0, 0, 0, 1, 1, 2, 3, 4, 5])
                self.cols = np.array([0, 1, 2, 1, 2, 2, 3, 4, 5])
        elif ss == 9:
            assert self.save_covariance == "diag"
            if allocate_space_for_direction:
                raise NotImplementedError("")
            else:
                self.desription = KalmanEstimates

    def get_description(self):
        return self.description

    def get_save_covariance(self):
        return self.save_covariance

    def covar_mat_to_covar_entries(self, M):
        if self.save_covariance == "diag":
            return np.diag(M).tolist()
        else:
            return np.asarray(M)[self.rows, self.cols].tolist()

    def covar_mats_to_covar_entries(self, Ms):
        result = []
        for (row, col) in zip(self.rows, self.cols):
            result.append(Ms[:, row, col])
        return result


class FilteredObservations(
    PT.IsDescription
):  # Not really "observations" but ML estimates
    obj_id = PT.UInt32Col(pos=0)
    frame = PT.Int64Col(pos=1)
    x = PT.Float32Col(pos=2)
    y = PT.Float32Col(pos=3)
    z = PT.Float32Col(pos=4)
    obs_2d_idx = PT.UInt64Col(pos=5)  # index into VLArray 'ML_estimates_2d_idxs'
    hz_line0 = PT.Float32Col(pos=6)
    hz_line1 = PT.Float32Col(pos=7)
    hz_line2 = PT.Float32Col(pos=8)
    hz_line3 = PT.Float32Col(pos=9)
    hz_line4 = PT.Float32Col(pos=10)
    hz_line5 = PT.Float32Col(pos=11)


ML_estimates_2d_idxs_type = PT.UInt16Atom


def convert_format(current_data, camn2cam_id, area_threshold=0.0, only_likely=False):
    """convert data from format used for Kalman tracker to hypothesis tester"""
    found_data_dict = {}
    first_idx_by_cam_id = {}
    for camn, stuff_list in current_data.items():
        if not len(stuff_list):
            # no data for this camera, continue
            continue
        for (pt_undistorted, projected_line) in stuff_list:
            if not np.isnan(pt_undistorted[0]):  # only use if point was found

                # perform area filtering
                area = pt_undistorted[PT_TUPLE_IDX_AREA]
                if area < area_threshold:
                    continue

                cam_id = camn2cam_id[camn]

                if only_likely:
                    # a quick gating based on image attributes.

                    pt_area = pt_undistorted[PT_TUPLE_IDX_AREA]
                    cur_val = pt_undistorted[PT_TUPLE_IDX_CUR_VAL_IDX]
                    mean_val = pt_undistorted[PT_TUPLE_IDX_MEAN_VAL_IDX]
                    sumsqf_val = pt_undistorted[PT_TUPLE_IDX_SUMSQF_VAL_IDX]

                    p_y_x = some_rough_negative_log_likelihood(
                        pt_area, cur_val, mean_val, sumsqf_val
                    )  # this could even depend on 3d geometry
                    if not np.isfinite(p_y_x):
                        continue

                found_data_dict[cam_id] = pt_undistorted[:9]
                first_idx_by_cam_id[cam_id] = pt_undistorted[PT_TUPLE_IDX_FRAME_PT_IDX]
                break  # algorithm only accepts 1 point per camera
    return found_data_dict, first_idx_by_cam_id



import numpy
import numpy as np


class EKF(object):
    """Extendend Kalman Filter

    This is only a "half" EKF - only the observation model is
    non-linear.  This non-linear function h() maps the a priori
    estimate of position into camera coordinates for N cameras.

    """

    def __init__(self, initial_x=None, initial_P=None):
        # These 2 attributes are the only state that changes during
        # filtering:
        self.xhat_k1 = initial_x  # a posteriori state estimate from step (k-1)
        self.P_k1 = initial_P  # a posteriori error estimate from step (k-1)

        self.ss = len(self.xhat_k1)

    def step1__calculate_a_priori(self, A, Q, isinitial=False):
        dot = numpy.dot  # shorthand
        ############################################
        #          update state-space

        # compute a priori estimate of statespace
        if not isinitial:
            AT = A.T
            xhatminus = dot(A, self.xhat_k1)
            # compute a priori estimate of errors
            Pminus = dot(dot(A, self.P_k1), AT) + Q
        else:
            xhatminus = self.xhat_k1
            Pminus = self.P_k1

        return xhatminus, Pminus

    def step2__calculate_a_posteriori(
        self, xhatminus, Pminus, y=None, hx=None, C=None, R=None, missing_data=False
    ):

        ss = len(xhatminus)  # state space size
        assert Pminus.shape == (ss, ss)

        if not missing_data:

            CT = C.T

            y = np.asanyarray(y)
            hx = np.asanyarray(hx)

            os = len(y)
            assert y.ndim == 1, "observations must be vector"
            assert hx.ndim == 1, "expected values must be vector"
            assert CT.shape == (ss, os)
            assert len(hx) == os
            assert R.shape == (os, os)

            dot = numpy.dot
            inv = numpy.linalg.inv

            ############################################
            #          incorporate observation

            # calculate a posteriori state estimate

            # calculate Kalman gain
            Knumerator = dot(Pminus, CT)
            Kdenominator = dot(dot(C, Pminus), CT) + R
            K = dot(Knumerator, inv(Kdenominator))  # Kalman gain

            residuals = y - hx
            xhat = xhatminus + dot(K, residuals)
            one_minus_KC = numpy.eye(ss) - dot(K, C)

            # compute a posteriori estimate of errors
            P = dot(one_minus_KC, Pminus)
        else:
            # no observation
            xhat = xhatminus
            P = Pminus

        # this step (k) becomes next step's prior (k-1)
        self.xhat_k1 = xhat
        self.P_k1 = P
        ##         if 1:
        ##             print 'xhat'
        ##             print xhat
        ##             print
        return xhat, P

    def step(self, A, Q):
        xhatminus, Pminus = self.step1__calculate_a_priori(A, Q, isinitial=False)
        return self.step2__calculate_a_posteriori(xhatminus, Pminus, missing_data=True)



"""

 1--
 ^  \---                      medium 1
 |       \----
 |            \---
 z1               \----
 |                     \---
 |                         \----
 V                              \---
 ====================================0====== interface
 ^                                    \
 |                                     \
 z2                                     \
 |      medium 2                         \
 v                                        \
 -                                         2

 |<--------------h1----------------->|
 |<--------------------h------------------>|
                                     |<-h2>|

h1 is the horizontal distance from point 1 to point 0.
h is the horitzontal distance from point 1 to point 2.
h2 is the horizontal distance from point 0 to point 2.

n1 is refractive index of medium 1
n2 is refractive index of medium 2

z1 is height of point 1
z2 is depth of point 2
"""
from sympy import *
from sympy.solvers.solvers import unrad # tested with sympy==0.7.3

h1 = Symbol('h1', real=True, positive=True)
h = Symbol('h', real=True, positive=True)
h2 = h-h1

n1 = Symbol('n1', real=True, positive=True)
n2 = Symbol('n2', real=True, positive=True)

z1 = Symbol('z1', real=True, positive=True)
z2 = Symbol('z2', real=True, positive=True)

duration = n1*sqrt( h1*h1 + z1*z1 ) + n2*sqrt(z2*z2 + h2*h2)

ddur_dh1 = diff(duration,h1)
eq, cov, dens = unrad(ddur_dh1)
poly = Poly(eq,h1)
print poly



#emacs, this is -*-Python-*- mode
#cython: language_level=2
cimport numpy as np

cdef class PinholeCameraModelWithJacobian:
    cdef double P00, P01, P02, P03
    cdef double P10, P11, P12, P13
    cdef double P20, P21, P22, P23
    cdef object pmat
    cdef void evaluate_jacobian_at_(self,double x,double y,double z,double w,
                                    double *ux, double *uy, double *uz, double *uw,
                                    double *vx, double *vy, double *vz, double *vw)
    cpdef evaluate_jacobian_at(self, np.ndarray[np.double_t, ndim=1] X)
    cpdef evaluate(self, np.ndarray[np.double_t, ndim=1] X)




# -*- coding: utf-8 -*-
#cython: language_level=2
cimport _Roots3And4
import _Roots3And4

"""
This code calculates rays according to [Fermat's principle of least
time](http://en.wikipedia.org/wiki/Fermat's_principle). Light
traveling from point 1 to point 2 (or vice-versa) takes the path of
least time.

 1--
    \---                      medium 1
        \----
             \---
                 \----
                      \---
                          \----
                               \---
 ====================================0====== interface
                                      \
                                       \
                                        \
      medium 2                           \
                                          \
                                           2


n1: refractive index of medium 1
n2: refractive index of medium 2
z1: height of point 1 (always positive)
z2: depth  of point 2 (always positive)
h1: horizontal distance between points 1,0 (always positive)
h2: horizontal distance between points 2,0 (always positive)
h:  horizontal distance between points 1,2 (always positive)

duration = n1*sqrt( h1*h1 + z1*z1 ) + n2*sqrt(z2*z2 + h2*h2)
ddur_dh1 = diff(duration,h1)
#print solve(ddur_dh1,dh1) # fails due to quartic polynomial
"""

cpdef double find_fastest_path_fermat(double n1,double n2,double z1,double h,double z2,double eps) except *:
    cdef double result
    cdef double a,b,c,d,e

    if z2==0.0:
        return h

    # see refraction_demo.py for factorization
    a = (n1**2 - n2**2)
    b = (-2*h*n1**2 + 2*h*n2**2)
    c = (h**2*n1**2 - h**2*n2**2 + n1**2*z2**2 - n2**2*z1**2)
    d = 2*h*n2**2*z1**2
    e = -h**2*n2**2*z1**2

    # According to
    #
    #  Glaeser, G., & Schröcker, H. (2000). Reflections on
    #  Refractions. JOURNAL FOR GEOMETRY AND GRAPHICS VOLUME, 4(1),
    #  1–18.
    #
    # we can choose the real root less than h. I (ADS) also found that
    # it should be non-negative.

    result = _Roots3And4.real_nonnegative_root_less_than(a,b,c,d,e,h, eps)
    return result



#emacs, this is -*-Python-*- mode
#cython: language_level=2

"""calculate jacobian matrices for pinhole camera model

see pinhole_jacobian_demo.py for the derivation of the math in this module.
"""
import numpy
import numpy as np
cimport numpy as np

def make_PinholeCameraModelWithJacobian(*args,**kw):
    return PinholeCameraModelWithJacobian(*args,**kw)

cdef class PinholeCameraModelWithJacobian:
    """Represent the 3D projection of a camera.

    Camera projection is a non-linear process where the 3D coordinate
    X = (x,y,z) is transformed to the 2D coordinate Y = (r,s) on the
    image plane by the function h:

                          Y = h(X)

    This class implements this function and allows one to evaluate the
    jacobian at various X, as well.

    """
    def __init__(self,P):
        self.pmat = P
        self.P00, self.P01, self.P02, self.P03 = P[0,:]
        self.P10, self.P11, self.P12, self.P13 = P[1,:]
        self.P20, self.P21, self.P22, self.P23 = P[2,:]

    def __reduce__(self):
        """this allows PinholeCameraModelWithJacobian to be pickled"""
        args = (self.pmat,)
        return (make_PinholeCameraModelWithJacobian, args)

    cdef void evaluate_jacobian_at_(self,double x,double y,double z,double w,
                                    double *ux, double *uy, double *uz, double *uw,
                                    double *vx, double *vy, double *vz, double *vw):
        """This is implemented for 3D homogeneous vector (x,y,z,w)"""

        #sympy.diff(u,x)
        ux[0]=-self.P20*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P00*x + self.P01*y + self.P02*z + self.P03*w) + self.P00/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(u,y)
        uy[0]=-self.P21*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P00*x + self.P01*y + self.P02*z + self.P03*w) + self.P01/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(u,z)
        uz[0]=-self.P22*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P00*x + self.P01*y + self.P02*z + self.P03*w) + self.P02/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(u,w)
        uw[0]=-self.P23*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P00*x + self.P01*y + self.P02*z + self.P03*w) + self.P03/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)

        #sympy.diff(v,x)
        vx[0]=-self.P20*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P10*x + self.P11*y + self.P12*z + self.P13*w) + self.P10/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(v,y)
        vy[0]=-self.P21*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P10*x + self.P11*y + self.P12*z + self.P13*w) + self.P11/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(v,z)
        vz[0]=-self.P22*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P10*x + self.P11*y + self.P12*z + self.P13*w) + self.P12/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
        #sympy.diff(v,w)
        vw[0]=-self.P23*(self.P20*x + self.P21*y + self.P22*z + self.P23*w)**(-2)*(self.P10*x + self.P11*y + self.P12*z + self.P13*w) + self.P13/(self.P20*x + self.P21*y + self.P22*z + self.P23*w)
#        return ux, uy, uz, uw, vx, vy, vz, vw
    def evaluate_jacobian_from_homogeneous(self,X):
        cdef double ux, uy, uz, uw
        cdef double vx, vy, vz, vw
        cdef double w

        x,y,z,w=X
        self.evaluate_jacobian_at_(x,y,z,w, &ux, &uy, &uz, &uw,
                                   &vx, &vy, &vz, &vw)
        J=numpy.array([[ux,uy,uz,uw],
                       [vx,vy,vz,vw]],
                      dtype=numpy.float64)
        return J
    cpdef evaluate_jacobian_at(self, np.ndarray[np.double_t, ndim=1] X): # tested on Cython 0.19.2
        cdef double ux, uy, uz, uw
        cdef double vx, vy, vz, vw
        cdef double w

        if len(X)==4:
            assert X[3]==1.0
        else:
            assert len(X)==3
        x,y,z=X[:3]
        w=1.0
        self.evaluate_jacobian_at_(x,y,z,w, &ux, &uy, &uz, &uw,
                                   &vx, &vy, &vz, &vw)
        J=numpy.array([[ux,uy,uz],
                       [vx,vy,vz]],
                      dtype=numpy.float64)
        return J
    cpdef evaluate(self, np.ndarray[np.double_t, ndim=1] X):
        """evaluate the non-linear function

        Y = h(X)
        """

        if len(X)==4:
            assert X[3]==1.0
        else:
            assert len(X)==3

        predicted3d_homogeneous = numpy.ones((4,1))
        predicted3d_homogeneous[:3,0] = X[:3]
        uvw = numpy.dot( self.pmat, predicted3d_homogeneous )
        Y = (uvw / uvw[2])[:2,0]
        return Y



from __future__ import print_function
import roslib
import roslib.packages
import rospkg

roslib.load_manifest("rospy")
roslib.load_manifest("rosgraph")
import rospy
import rospy.names
import rosgraph.masterapi

import os.path
import string
import sys
import errno
from urlparse import urlparse
import socket


class Log:
    to_console = True
    force_output = None

    def __init__(self, to_ros=False):
        self.to_ros = to_ros

    def _write(self, x, f):
        _file = self.force_output if self.force_output != None else f
        if x:
            while 1:
                try:
                    _file.write(x)
                    if not x.endswith("\n"):
                        _file.write("\n")
                    break
                except IOError as err:
                    if err.args[0] == errno.EINTR:  # interrupted system call
                        continue
        while 1:
            try:
                _file.flush()
                break
            except IOError as err:
                if err.args[0] == errno.EINTR:
                    continue

    def info(self, x):
        if self.to_ros:
            # loginfo prints to stdout
            rospy.loginfo(x)
            return
        if self.to_console:
            self._write(x, sys.stdout)

    def debug(self, x):
        if self.to_ros:
            rospy.logdebug(x)
        if self.to_console:
            self._write(x, sys.stdout)

    def warn(self, x):
        if self.to_ros:
            rospy.logwarn(x)
            # logwarn prints to stderr
            return
        if self.to_console:
            self._write(x, sys.stderr)

    def fatal(self, x):
        if self.to_ros:
            rospy.logfatal(x)
            # logwarn prints to stderr
            return
        if self.to_console:
            self._write(x, sys.stderr)


# FROM motmot_ros_utils.roslib.io
def decode_url(url, required=False):
    """
    Example URL syntax:
        file:///full/path/to/local/file.yaml
        file:///full/path/to/videre/file.ini
        package://camera_info_manager/tests/test_calibration.yaml
        package://ros_package_name/calibrations/camera3.yaml
        /full/path
        ~/home/path

    The file: URL specifies a full path name in the local system. The package: URL is handled
    the same as file:, except the path name is resolved relative to the location of the named ROS
    package, which must be reachable via $ROS_PACKAGE_PATH.

    The following variables can be replaced

        ${NODE_NAME} resolved to the current node name.
        ${ROS_HOME} resolved to the $ROS_HOME environment variable if defined, ~/.ros if not.

    XXXX: Use the ros resource_retriever class here one day, when it has python bindings and
          variable substitution
    """
    if url.startswith("~"):
        url = os.path.expanduser(url)
    elif url.startswith("file://"):
        url = url[7:]
    elif url.startswith("package://"):
        package, fragment = url[10:].split("/", 1)
        url = os.path.join(
            roslib.packages.get_pkg_dir(package, required=False), fragment
        )

    nodename = rospy.names.get_name()
    if nodename:
        # rospy returns rully resolved name, so remove first char (~,/) and
        # return only the first fragment
        nodename = nodename[1:].split(" ")[-1]
    else:
        nodename = ""

    url = string.Template(url).safe_substitute(
        NODE_NAME=nodename, ROS_HOME=rospkg.get_ros_home()
    )

    url = os.path.abspath(url)
    if required and not os.path.exists(url):
        raise Exception("resource not found")

    return url


def get_node_hostname(node_name):
    ID = rospy.get_name()  # my own name
    master = rosgraph.masterapi.Master(ID)

    # Get URI of other node's XMLRPC server.
    uri = master.lookupNode(node_name)

    # Parse the URI to find the hostname of the server.
    x = urlparse(uri)
    if "[" in x.netloc:
        # must be IPv6 according to RFC3986
        if not "]" in x.netloc:
            raise ValueError("invalid IPv6 address")
        start_idx = x.netloc.index("[")
        stop_idx = x.netloc.index("]")
        assert stop_idx > start_idx
        hostname = x.netloc[start_idx + 1 : stop_idx]
    elif ":" in x.netloc:
        hostname = x.netloc[: x.netloc.rindex(":")]
    else:
        hostname = x.netloc

    return hostname



#emacs, this is -*-Python-*- mode
#cython: language_level=2
import numpy
cimport flydra_core._fastgeom as _fastgeom

def line_fit_2d(line_2d, mu, sigma_inv):
    """find the point closest to mu on the line according to Mahalanois distance"""
    cdef double x0, x1, S00, S01, S10, S11, a,b,c,d

    a,c,b,d = line_2d # parametric representation of line passing through (a,c) in direction (b,d)

    S00 = sigma_inv[0,0]
    S01 = sigma_inv[0,1]
    S10 = sigma_inv[1,0]
    S11 = sigma_inv[1,1]

    x0 = mu[0]
    x1 = mu[1]

    # this was generated with sympy (see mahal.history)
    s = 1.0/(-2*S01*b*d - 2*S10*b*d - 2*S00*b**2 - 2*S11*d**2)*(S01*a*d + S01*b*c + S10*a*d + S10*b*c - S01*b*x1 - S01*d*x0 - S10*b*x1 - S10*d*x0 - 2*S00*b*x0 - 2*S11*d*x1 + 2*S00*a*b + 2*S11*c*d)

    loc2d = (a+s*b,c+s*d)
    return loc2d

cpdef _fastgeom.ThreeTuple line_fit_3d(_fastgeom.PlueckerLine line_pluecker, _fastgeom.ThreeTuple mu, sigma_inv):
    """find the point closest to mu on the line according to Mahalanobis distance"""

    # The idea -- Because the Mahalanobis distance function is convex,
    # find the inflection point (where the derivative is zero).

    cdef double x0, x1, x2, S00, S01, S02, S10, S11, S12, S20, S21, S22, a,b,c,d,e,f

    cdef _fastgeom.ThreeTuple lu, lv

    lu = line_pluecker.u
    lv = line_pluecker.v

    a,b,c = lu.a, lu.b, lu.c
    d,e,f = lv.a, lv.b, lv.c

    # slice into potentially bigger arrays
    x0 = mu.a
    x1 = mu.b
    x2 = mu.c

    S00,S01,S02 = sigma_inv[0,:3]
    S10,S11,S12 = sigma_inv[1,:3]
    S20,S21,S22 = sigma_inv[2,:3]

    # this was generated with sympy (see mahal.history)
    s = 1.0/(-2*S01*d*e - 2*S02*d*f - 2*S10*d*e - 2*S12*e*f - 2*S20*d*f - 2*S21*e*f - 2*S00*d**2 - 2*S11*e**2 - 2*S22*f**2)*(S01*a*e + S01*b*d + S02*a*f + S02*c*d + S10*a*e + S10*b*d + S12*b*f + S12*c*e + S20*a*f + S20*c*d + S21*b*f + S21*c*e - S01*d*x1 - S01*e*x0 - S02*d*x2 - S02*f*x0 - S10*d*x1 - S10*e*x0 - S12*e*x2 - S12*f*x1 - S20*d*x2 - S20*f*x0 - S21*e*x2 - S21*f*x1 - 2*S00*d*x0 - 2*S11*e*x1 - 2*S22*f*x2 + 2*S00*a*d + 2*S11*b*e + 2*S22*c*f)

    cdef _fastgeom.ThreeTuple loc3d = _fastgeom.ThreeTuple((a+s*d, b+s*e, c+s*f))
    return loc3d

cpdef double dist2(_fastgeom.ThreeTuple x,_fastgeom.ThreeTuple mu,sigma_inv):
    """return the squared Mahalanobis distance"""
    assert sigma_inv.shape==(3,3)
    cdef _fastgeom.ThreeTuple r = x-mu
    sub = numpy.array((r.a, r.b, r.c))
    cdef double result = numpy.dot( sub.T, numpy.dot( sigma_inv, sub ) )
    return result



from __future__ import print_function
import numpy as np
import adskalman.adskalman as adskalman


def running_average(src, min_dist):
    if len(src) < min_dist:
        raise ValueError("the input sequence is too short")
    min_dist = 5
    accum = []
    for i in range(min_dist):
        accum.append(src[i : (len(src) - min_dist + i + 1)])
    accum = np.array(accum)
    near = np.mean(accum, axis=0)
    return near


def test_running_average():
    a = [0, 0, 0, 0, 0, 5, 5, 5, 5, 5]
    b = [0, 1, 2, 3, 4, 5]
    bex = running_average(a, 5)
    assert np.allclose(bex, b)

    a = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1], dtype=np.bool)
    b = [0, 0.2, 0.4, 0.6, 0.8, 1.0]
    bex = running_average(a, 5)
    assert np.allclose(bex, b)

    try:
        running_average([1, 2], 5)
    except ValueError:
        pass
    else:
        raise ValueError("failed to abort on short input")


def ori_smooth(directions, frames_per_second=None, return_missing=False):
    """smooth orientations using an RTS smoother

    This treats orientations as XYZ positions
    """
    dt = 1.0 / frames_per_second

    A = np.array(
        [
            [1, 0, 0, dt, 0, 0],
            [0, 1, 0, 0, dt, 0],
            [0, 0, 1, 0, 0, dt],
            [0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 1],
        ]
    )
    Q = 1.0 * np.eye(6)

    C = np.array([[1, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0]])
    R = 30.0 * np.eye(3)

    init_x = np.hstack((directions[0, :], np.zeros((3,))))
    assert not np.any(np.isnan(init_x)), "cannot start with missing orientation"
    init_V = 2 * Q
    y = directions
    dirsmooth, V = adskalman.kalman_smoother(y, A, C, Q, R, init_x, init_V)
    dirsmooth = dirsmooth[:, :3]  # take only position information
    dirsmooth_missing = np.array(dirsmooth, copy=True)

    # remove results too distant from observations
    avlen = 5
    if len(y) >= avlen:
        good = ~np.isnan(y[:, 0])
        near_good = running_average(good, avlen)
        pad = avlen // 2
        near_good = np.hstack((np.zeros((pad,)), near_good, np.zeros((pad,))))

        good_cond = near_good > 0.2
        bad_cond = ~good_cond

        if bad_cond.shape != dirsmooth[:, 0].shape:
            print("xxxx")
            print(bad_cond.shape)
            print(dirsmooth[:, 0].shape)
            print(directions.shape)
        dirsmooth[bad_cond, :] = np.nan

    # normalize lengths to unit vectors
    np.sum(dirsmooth ** 2, axis=1)
    dirsmooth = (dirsmooth.T / np.sqrt(np.sum(dirsmooth ** 2, axis=1))).T

    if return_missing:
        return dirsmooth, dirsmooth_missing
    else:
        return dirsmooth


def test_ori_smooth():
    directions = np.zeros([300, 3])
    directions[:, 0] = 1.0
    smoothed = ori_smooth(directions, frames_per_second=1.0)



MINIMUM_ECCENTRICITY = 1.4  # threshold to fit line

# epsilon for numerical precision when computing quartic roots in
# refractive boundary code
WATER_ROOTS_EPS = 1e-5

timestamp_echo_listener_port = 9992  # on cameras
timestamp_echo_fmt1 = "<d"
timestamp_echo_fmt_diff = "<d"
timestamp_echo_gatherer_port = 9993  # on MainBrain
timestamp_echo_fmt2 = "<dd"

cam_id_count = 30
recv_pt_fmt = "<dddddBBdd"
recv_pt_header_fmt = "<%dpddliI" % (cam_id_count,)

emulated_camera_control = 9645

sync_duration = 2.0  # seconds to pause triggering to cause synchronization
near_inf = 9.999999e20



"""linear and non-linear dynamic models for different animals"""

import numpy
import warnings
import re

DEFAULT_MODEL = "mamarama, units: mm"


def _get_decreasing_accel_model(dt=None):
    """get linear dynamical system matrices A and C

    dt is the time-step in seconds
    """
    # distance units are in m
    # time units are in sec
    # thus, velocity is (m/sec)

    ss = 9  # length of state vector (state size)
    os = 3  # length of observation vector (observation size)

    half_dt2 = 0.5 * dt ** 2
    ad = 0.1  # acceleration decay

    # state vector describes a particle in brownian motion
    # [ x y z xvel yvel zvel xaccel yaccel zaccel]

    # process update matrix (time evolution update matrix)
    A = numpy.array(
        [
            [1.0, 0.0, 0.0, dt, 0.0, 0.0, half_dt2, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, dt, 0.0, 0.0, half_dt2, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, dt, 0.0, 0.0, half_dt2],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, dt, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, dt, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, dt],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ad, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ad, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ad],
        ]
    )
    A_model_name = "decreasing_accel"

    # measurement prediction matrix
    C = numpy.zeros((os, ss))
    C[:os, :os] = numpy.eye(os)  # directly measure x,y,z positions

    ### process covariance
    ##Q = numpy.zeros((ss,ss))
    ##for i in range(6,9):
    ##    Q[i,i] = 10.0 # acceleration noise (near (3.16m*sec**-2)**2)

    ### measurement noise covariance matrix
    ##R = 1e-6*numpy.eye(os) # (1mm)**2 = (0.001m)**2
    model = {
        "A": A,
        "A_model_name": A_model_name,
        "C": C,
        "ss": ss,
        "os": os,
        "dt": dt,
    }
    return model


def _get_fixed_vel_model(dt=None):
    """get linear dynamical system matrices A and C

    dt is the time-step in seconds
    """
    # distance units are in m
    # time units are in sec
    # thus, velocity is (m/sec)

    ss = 6  # length of state vector (state size)
    os = 3  # length of observation vector (observation size)

    # state vector describes a particle in brownian motion
    # [ x y z xvel yvel zvel]

    # process update matrix (time evolution update matrix)
    A = numpy.array(
        [
            [1.0, 0.0, 0.0, dt, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, dt, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, dt],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        ]
    )
    A_model_name = "fixed_vel"

    # measurement prediction matrix
    C = numpy.zeros((os, ss))
    C[:os, :os] = numpy.eye(os)  # directly measure x,y,z positions

    model = {
        "A": A,
        "A_model_name": A_model_name,
        "C": C,
        "ss": ss,
        "os": os,
        "dt": dt,
    }
    return model


def get_dynamic_model_dict(*args, **kw):
    warnings.warn(
        'DeprecationWarning: call "get_kalman_model()", not old "get_dynamic_model_dict()"'
    )
    return create_dynamic_model_dict(*args, **kw)


def create_dynamic_model_dict(dt=None, disable_warning=False):
    """get linear dynamical system matrices

    dt is the time-step in seconds
    """
    if not disable_warning:
        warnings.warn(
            'using deprecated function "create_dynamic_model_dict()". Use "get_kalman_model()"',
            DeprecationWarning,
            stacklevel=2,
        )
    dynamic_models = {}

    ######################################
    # 'hbird, units: mm':
    # process covariance

    # WARNING: these parameters haven't been tested since the
    # consolidation of the flydra calibration stuff in July-August
    # 2012.

    base_model_dict = _get_fixed_vel_model(dt)
    ss = base_model_dict["ss"]
    os = base_model_dict["os"]

    Q = numpy.zeros((ss, ss))
    for i in range(0, 3):
        Q[i, i] = (0.04) ** 2
    for i in range(3, 6):
        Q[i, i] = (0.4) ** 2

    # measurement noise covariance matrix
    R = 1e-2 * numpy.eye(os)

    newdict = dict(
        # data association parameters
        # birth model
        hypothesis_test_max_acceptable_error=50.0,
        min_dist_to_believe_new_meters=0.2,  # 20 cm
        min_dist_to_believe_new_sigma=3.0,
        initial_position_covariance_estimate=1e-2,
        initial_velocity_covariance_estimate=10,
        # death model
        max_variance_dist_meters=0.08,
        max_frames_skipped=10,
        # kalman filter parameters
        Q=Q,
        R=R,
    )
    newdict.update(base_model_dict)
    dynamic_models["hbird, units: mm"] = newdict

    ######################################

    # 'mamarama, units: mm':
    # process covariance

    base_model_dict = _get_fixed_vel_model(dt)
    ss = base_model_dict["ss"]
    os = base_model_dict["os"]

    if 1:
        # this form is after N. Shimkin's lecture notes in
        # Estimation and Identification in Dynamical Systems
        # http://webee.technion.ac.il/people/shimkin/Estimation09/ch8_target.pdf
        assert ss == 6
        T33 = dt ** 3 / 3.0
        T22 = dt ** 2 / 2.0
        T = dt
        Q = numpy.array(
            [
                [T33, 0, 0, T22, 0, 0],
                [0, T33, 0, 0, T22, 0],
                [0, 0, T33, 0, 0, T22],
                [T22, 0, 0, T, 0, 0],
                [0, T22, 0, 0, T, 0],
                [0, 0, T22, 0, 0, T],
            ]
        )
    # the scale of the covariance
    Q = 0.01 * Q

    # measurement noise covariance matrix
    R = 1e-7 * numpy.eye(os)

    newdict = dict(
        # data association parameters
        # birth model
        hypothesis_test_max_acceptable_error=5.0,
        min_dist_to_believe_new_meters=0.02,
        min_dist_to_believe_new_sigma=10.0,
        initial_position_covariance_estimate=1e-3,
        initial_velocity_covariance_estimate=10,
        # death model
        max_variance_dist_meters=0.25,
        max_frames_skipped=10,
        # kalman filter parameters
        Q=Q,
        R=R,
    )
    newdict.update(base_model_dict)
    dynamic_models["mamarama, units: mm"] = newdict

    ######################################
    # 'fishbowl40':
    # process covariance

    base_model_dict = _get_fixed_vel_model(dt)
    ss = base_model_dict["ss"]
    os = base_model_dict["os"]

    if 1:
        # this form is after N. Shimkin's lecture notes in
        # Estimation and Identification in Dynamical Systems
        # http://webee.technion.ac.il/people/shimkin/Estimation09/ch8_target.pdf
        assert ss == 6
        T33 = dt ** 3 / 3.0
        T22 = dt ** 2 / 2.0
        T = dt
        Q = numpy.array(
            [
                [T33, 0, 0, T22, 0, 0],
                [0, T33, 0, 0, T22, 0],
                [0, 0, T33, 0, 0, T22],
                [T22, 0, 0, T, 0, 0],
                [0, T22, 0, 0, T, 0],
                [0, 0, T22, 0, 0, T],
            ]
        )
    # the scale of the covariance
    Q = 0.01 * Q

    # measurement noise covariance matrix
    R = 1e-7 * numpy.eye(os)

    newdict = dict(
        # data association parameters
        # birth model
        hypothesis_test_max_acceptable_error=5.0,  # Big Fish 5    Smallfish 5
        min_dist_to_believe_new_meters=0.05,  # Big Fish 0.05   Smallfish 0.02
        min_dist_to_believe_new_sigma=10.0,
        initial_position_covariance_estimate=1e-3,
        initial_velocity_covariance_estimate=10,
        # death model
        max_variance_dist_meters=0.125,
        max_frames_skipped=30,
        # kalman filter parameters
        Q=Q,
        R=R,
    )
    newdict.update(base_model_dict)
    dynamic_models["fishbowl40"] = newdict

    ######################################

    # 'hydra, units: m':
    # process covariance

    # WARNING: these parameters haven't been tested since the
    # consolidation of the flydra calibration stuff in July-August
    # 2012.

    base_model_dict = _get_fixed_vel_model(dt)
    ss = base_model_dict["ss"]
    os = base_model_dict["os"]

    Q = numpy.zeros((ss, ss))
    for i in range(0, 3):
        Q[i, i] = (0.01) ** 2
    for i in range(3, 6):
        Q[i, i] = (0.5) ** 2

    Q = Q * 1000 ** 2  # convert to meters

    # measurement noise covariance matrix
    R = 1e-3 * numpy.eye(os)

    newdict = dict(
        # data association parameters
        # birth model
        hypothesis_test_max_acceptable_error=50.0,
        min_dist_to_believe_new_meters=0.08,  # 8 cm
        min_dist_to_believe_new_sigma=3.0,
        initial_position_covariance_estimate=1e-6,
        initial_velocity_covariance_estimate=1,
        # death model
        max_variance_dist_meters=0.08,
        max_frames_skipped=10,
        # kalman filter parameters
        Q=Q,
        R=R,
    )
    newdict.update(base_model_dict)
    dynamic_models["hydra, units: m"] = newdict

    ## ##################################################

    return dynamic_models


class EKFAllParams(dict):
    """abstract base class hold all parameters for data association and EK filtering"""

    def __init__(self):
        self["ss"] = 6
        self["isEKF"] = True


class MamaramaMMEKFAllParams(EKFAllParams):
    """Drosophila non-linear dynamic model for EKF"""

    def __init__(self, dt=None):
        super(MamaramaMMEKFAllParams, self).__init__()
        assert dt is not None
        linear_dict = get_kalman_model(name="mamarama, units: mm", dt=dt)

        # update some parameters from linear model
        for key in [
            "initial_position_covariance_estimate",
            "max_frames_skipped",
            "A",
            "Q",
            "dt",
            "hypothesis_test_max_acceptable_error",
            "min_dist_to_believe_new_meters",
            "min_dist_to_believe_new_sigma",
            "max_variance_dist_meters",
        ]:
            self[key] = linear_dict[key]
        self["ekf_observation_covariance_pixels"] = numpy.array(
            [[1.0, 0.0], [0.0, 1.0]], dtype=numpy.float64
        )
        self[
            "distorted_pixel_euclidian_distance_accept"
        ] = 20.0  # distance in the raw image plane (i.e. before radial undistortion)


class Fishbowl40EKFAllParams(EKFAllParams):
    def __init__(self, dt=None):
        super(Fishbowl40EKFAllParams, self).__init__()
        assert dt is not None
        linear_dict = get_kalman_model(name="fishbowl40", dt=dt)

        # update some parameters from linear model
        for key in [
            "initial_position_covariance_estimate",
            "max_frames_skipped",
            "A",
            "Q",
            "dt",
            "hypothesis_test_max_acceptable_error",
            "min_dist_to_believe_new_meters",
            "min_dist_to_believe_new_sigma",
            "max_variance_dist_meters",
        ]:
            self[key] = linear_dict[key]
        self["ekf_observation_covariance_pixels"] = numpy.array(
            [[1.0, 0.0], [0.0, 1.0]], dtype=numpy.float64
        )
        self[
            "distorted_pixel_euclidian_distance_accept"
        ] = 20.0  # distance in the raw image plane (i.e. before radial undistortion)


class HydraMEKFAllParams(EKFAllParams):
    # WARNING: these parameters haven't been tested since the
    # consolidation of the flydra calibration stuff in July-August
    # 2012.

    def __init__(self, dt=None):
        super(HydraMEKFAllParams, self).__init__()
        assert dt is not None
        linear_dict = get_kalman_model(name="hydra, units: m", dt=dt)

        # update some parameters from linear model
        for key in [
            "initial_position_covariance_estimate",
            "max_frames_skipped",
            "A",
            "Q",
            "dt",
        ]:
            self[key] = linear_dict[key]
        self["ekf_observation_covariance_pixels"] = numpy.array(
            [[1.0, 0.0], [0.0, 1.0]], dtype=numpy.float64
        )

        self["Q"] = self["Q"] / (1000 ** 2)
        self["min_dist_to_believe_new_meters"] = 0.2
        self["min_dist_to_believe_new_sigma"] = 10.0

        self[
            "distorted_pixel_euclidian_distance_accept"
        ] = 20.0  # distance in the raw image plane (i.e. before radial undistortion)

        if 0:
            # restrictive (better for e.g. making new calibration)
            self["max_variance_dist_meters"] = 0.25
        else:
            # loosy-goosy
            self["max_variance_dist_meters"] = 2  # let grow huge


class HbirdEKFAllParams(EKFAllParams):
    # WARNING: these parameters haven't been tested since the
    # consolidation of the flydra calibration stuff in July-August
    # 2012.

    def __init__(self, dt=None):
        super(HbirdEKFAllParams, self).__init__()
        assert dt is not None
        linear_dict = get_kalman_model(name="mamarama, units: mm", dt=dt)

        # update some parameters from linear model
        for key in [
            "initial_position_covariance_estimate",
            "max_frames_skipped",
            "A",
            "Q",
            "dt",
            "hypothesis_test_max_acceptable_error",
            "min_dist_to_believe_new_meters",
            "min_dist_to_believe_new_sigma",
            "max_variance_dist_meters",
        ]:
            self[key] = linear_dict[key]
        self["ekf_observation_covariance_pixels"] = numpy.array(
            [[15.0, 0.0], [0.0, 15.0]], dtype=numpy.float64
        )
        # distance in the raw image plane (i.e. before radial undistortion)
        self["distorted_pixel_euclidian_distance_accept"] = 15.0


ekf_models = {
    "EKF mamarama, units: mm": MamaramaMMEKFAllParams,
    "EKF fishbowl40": Fishbowl40EKFAllParams,
    "EKF hydra, units: m": HydraMEKFAllParams,
    "EKF hbird, units: mm": HbirdEKFAllParams,
}


def get_model_names(ekf_ok=True):
    """get names of available Kalman models"""
    model_dict = create_dynamic_model_dict(dt=0.01, disable_warning=True)
    valid_names = model_dict.keys()
    if ekf_ok:
        valid_names += ekf_models.keys()
    valid_names.sort()
    return valid_names


def get_kalman_model(name=None, dt=None):
    """create a return a Kalman model given a name and timestamp"""
    if name is None:
        raise ValueError("cannot get Kalman model unless name is specified")
    if name.startswith("EKF"):
        if name in ekf_models:
            klass = ekf_models[name]
            kalman_model = klass(dt=dt)
        else:
            raise KeyError("unknown EKF model: %s" % str(name))
    else:
        if name.startswith("fixed_vel_model"):
            # specify fixed_vel_model parameters as string

            # This is modified from Python Library ref 4.2.5:
            floatre = "([-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?)"

            rexp = (
                r"fixed_vel_model\(posQ=%s,velQ=%s,scalarR=%s,init_posV=%s,init_velV=%s\)"
                % (floatre, floatre, floatre, floatre, floatre)
            )
            matchobj = re.match(rexp, name)
            posQ, velQ, scalarR, init_posV, init_velV = map(float, matchobj.groups())

            base_model_dict = _get_fixed_vel_model(dt)
            ss = base_model_dict["ss"]
            os = base_model_dict["os"]

            Q = numpy.zeros((ss, ss))
            for i in range(0, 3):
                Q[i, i] = posQ
            for i in range(3, 6):
                Q[i, i] = velQ

            # measurement noise covariance matrix
            R = scalarR * numpy.eye(os)

            kalman_model = dict(
                Q=Q,
                R=R,
                initial_position_covariance_estimate=init_posV,
                initial_velocity_covariance_estimate=init_velV,
            )
            kalman_model.update(base_model_dict)
            return kalman_model
        model_dict = create_dynamic_model_dict(dt=dt, disable_warning=True)
        try:
            kalman_model = model_dict[name]
        except KeyError:
            valid_names = get_model_names()
            raise KeyError(
                "'%s', valid model names: %s"
                % (str(name), ", ".join(map(repr, valid_names)))
            )
    return kalman_model



from setuptools import setup, find_packages
from distutils.core import Extension  # actually monkey-patched by setuptools
from Cython.Build import cythonize

import numpy as np

from io import open
from os import path

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

ext_modules = []

ext_modules.append(
    Extension(
        name="flydra_core._reconstruct_utils",
        sources=["flydra_core/_reconstruct_utils.pyx"],
        include_dirs=[np.get_include()],
    )
)

ext_modules.append(
    Extension(
        name="flydra_core._pmat_jacobian",
        sources=["flydra_core/_pmat_jacobian.pyx"],
        include_dirs=[np.get_include()],
    )
)

ext_modules.append(
    Extension(
        name="flydra_core._pmat_jacobian_water",
        sources=["flydra_core/_pmat_jacobian_water.pyx"],
        include_dirs=[np.get_include()],
    )
)

ext_modules.append(
    Extension(
        name="flydra_core._flydra_tracked_object",
        sources=["flydra_core/_flydra_tracked_object.pyx"],
        include_dirs=[np.get_include()],
    )
)

ext_modules.append(
    Extension(name="flydra_core._mahalanobis", sources=["flydra_core/_mahalanobis.pyx"])
)

ext_modules.append(
    Extension(name="flydra_core._fastgeom", sources=["flydra_core/_fastgeom.pyx"])
)

ext_modules.append(
    Extension(
        name="flydra_core._Roots3And4",
        sources=["flydra_core/_Roots3And4.pyx", "flydra_core/Roots3And4.c",],
    )
)

ext_modules.append(
    Extension(name="flydra_core._refraction", sources=["flydra_core/_refraction.pyx"])
)

setup(
    name="flydra_core",
    version="0.7.14",  # keep in sync with flydra_core/version.py
    author="Andrew Straw",
    author_email="strawman@astraw.com",
    url="https://github.com/strawlab/flydra",
    description="flydra mainbrain and core lib",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    test_suite="nose.collector",
    ext_modules=cythonize(ext_modules),
    setup_requires=["setuptools_git >= 0.3",],
    entry_points={
        "console_scripts": [
            # camera calibration
            "flydra_analysis_align_calibration = flydra_core.reconstruct:align_calibration",
            "flydra_analysis_print_cam_centers = flydra_core.reconstruct:print_cam_centers",
            "flydra_analysis_flip_calibration = flydra_core.reconstruct:flip_calibration",
        ],
    },
    package_data={"flydra_core": ["flydra_server_art.png", "sample_calibration/*",],},
)



#cython: language_level=2
cimport _fastgeom

cpdef _fastgeom.ThreeTuple line_fit_3d(_fastgeom.PlueckerLine line_pluecker, _fastgeom.ThreeTuple mu, sigma_inv)
cpdef double dist2(_fastgeom.ThreeTuple x,_fastgeom.ThreeTuple mu,sigma_inv)



import sympy

S00=sympy.Symbol('S00')
S01=sympy.Symbol('S01')
S02=sympy.Symbol('S02')
S10=sympy.Symbol('S10')
S11=sympy.Symbol('S11')
S12=sympy.Symbol('S12')
S20=sympy.Symbol('S20')
S21=sympy.Symbol('S21')
S22=sympy.Symbol('S22')

x0=sympy.Symbol('x0')
x1=sympy.Symbol('x1')
x2=sympy.Symbol('x2')
y0=sympy.Symbol('y0')
y1=sympy.Symbol('y1')
y2=sympy.Symbol('y2')

post=sympy.Matrix(([y0-x0],[y1-x1],[y2-x2]))
S=sympy.Matrix(([S00,S01,S02],[S10,S11,S12],[S20,S21,S22]))
pre=post.T

maha3=(pre*S*post)[0,0]
(a,b,c,d,e,f,s)=[sympy.Symbol(x) for x in ('abcdefs')]
tmp = maha3
tmp = tmp.subs(y0,'a+s*d')
tmp = tmp.subs(y1,'b+s*e')
tmp = tmp.subs(y2,'c+s*f')
print 'solving with sympy',sympy.__version__
result = sympy.solve(sympy.diff(tmp,s),s)
print result



from __future__ import print_function
import flydra_core.data_descriptions as data_descriptions
import collections

from flydra_core._flydra_tracked_object import TrackedObject

__all__ = ["TrackedObject", "Tracker"]


class Tracker:
    """
    Handle multiple tracked objects using TrackedObject instances.

    This class keeps a list of objects currently being tracked. It
    also keeps a couple other lists for dealing with cases when the
    tracked objects are no longer 'live'.

    """

    def __init__(
        self,
        reconstructor,
        kalman_model=None,
        save_all_data=False,
        area_threshold=0,
        area_threshold_for_orientation=0.0,
        disable_image_stat_gating=False,
        orientation_consensus=0,
        fake_timestamp=None,
    ):
        """

        arguments
        =========
        reconstructor - reconstructor instance
        kalman_model - dictionary of Kalman filter parameters
        area_threshold - minimum area to consider for tracking use

        """
        self.area_threshold = area_threshold
        self.area_threshold_for_orientation = area_threshold_for_orientation
        self.save_all_data = save_all_data
        self.reconstructor = reconstructor
        self.live_tracked_objects = []
        self.dead_tracked_objects = []  # save for getting data out
        self.kill_tracker_callbacks = []
        self.flushed_callbacks = []
        self.disable_image_stat_gating = disable_image_stat_gating
        self.orientation_consensus = orientation_consensus
        self.fake_timestamp = fake_timestamp
        self.cur_obj_id = 1

        if kalman_model is None:
            raise ValueError("must specify kalman_model")
        self.kalman_model = kalman_model

    def how_many_are_living(self):
        # XXX should we check .kill_me attribute on them?
        return len(self.live_tracked_objects)

    def get_most_recent_data(self):
        results = [tro.get_most_recent_data() for tro in self.live_tracked_objects]
        return results

    def debug_info(level):
        _ = [tro.debug_info(level=level) for tro in self.live_tracked_objects]

    def is_believably_new(self, Xmm, debug=0):

        """Sometimes the Kalman tracker will not gobble all the points
        it should. This still prevents spawning a new Kalman
        tracker."""

        believably_new = True
        X = Xmm
        min_dist_to_believe_new_meters = self.kalman_model[
            "min_dist_to_believe_new_meters"
        ]
        min_dist_to_believe_new_nsigma = self.kalman_model[
            "min_dist_to_believe_new_sigma"
        ]
        results = [tro.get_distance_and_nsigma(X) for tro in self.live_tracked_objects]
        for (dist_meters, dist_nsigma) in results:
            if debug > 5:
                print("distance in meters, nsigma:", dist_meters, dist_nsigma)
            if (dist_nsigma < min_dist_to_believe_new_nsigma) or (
                dist_meters < min_dist_to_believe_new_meters
            ):
                believably_new = False
                break
        return believably_new

    def remove_duplicate_detections(self, frame, input_data_dict):
        """remove points that are close to current objects being tracked"""

        PT_TUPLE_IDX_FRAME_PT_IDX = data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX

        (test_frame, bad_camn_pts) = self.last_close_camn_pt_idxs
        assert test_frame == frame

        all_bad_pts = collections.defaultdict(set)
        for camn, ptnum in bad_camn_pts:
            all_bad_pts[camn].add(ptnum)

        output_data_dict = collections.defaultdict(list)
        for camn, camn_list in input_data_dict.items():
            bad_pts = all_bad_pts[camn]
            for element in camn_list:
                pt = element[0]
                ptnum = pt[PT_TUPLE_IDX_FRAME_PT_IDX]
                if ptnum not in bad_pts:
                    output_data_dict[camn].append(element)

        return output_data_dict

    def calculate_a_posteriori_estimates(self, frame, data_dict, camn2cam_id, debug2=0):
        # Allow earlier tracked objects to take all the data they
        # want.

        if debug2 > 1:
            print(self, "gobbling all data for frame %d" % (frame,))

        all_to_gobble = []
        best_by_hash = {}
        to_rewind = []
        # I could parallelize this========================================
        # this is map:
        results = [
            tro.calculate_a_posteriori_estimate(
                frame, data_dict, camn2cam_id, debug1=debug2
            )
            for tro in self.live_tracked_objects
        ]

        # this is reduce:
        all_close_camn_pt_idxs = []
        for idx, result in enumerate(results):
            (used_camns_and_idxs, obs2d_hash, Pmean, close_camn_pt_idxs) = result

            # Two similar lists -- lists of points that will be
            # removed from further consideration. "Gobbling" prevents
            # another object from using it if all the data were in
            # common. Removal of "close", probable duplicate
            # detections, does not remove consideration from
            # pre-existing objects, but will prevent birth of new
            # targets.

            # This works on entire group of 2D
            # observations. (I.e. individual 2D observations can support
            # multiple 3D targets. Only the entire set of 2D
            # observations, given by the obs2d_hash, cannot be shared.)

            all_to_gobble.extend(used_camns_and_idxs)
            all_close_camn_pt_idxs.extend(close_camn_pt_idxs)

            if obs2d_hash is not None:
                if obs2d_hash in best_by_hash:
                    (best_idx, best_Pmean) = best_by_hash[obs2d_hash]
                    if Pmean < best_Pmean:
                        # new value is better than previous best
                        best_by_hash[obs2d_hash] = (idx, Pmean)
                        to_rewind.append(best_idx)
                    else:
                        # old value is still best
                        to_rewind.append(idx)
                else:
                    best_by_hash[obs2d_hash] = (idx, Pmean)
        self.last_close_camn_pt_idxs = (frame, all_close_camn_pt_idxs)

        # End  ================================================================

        if len(all_to_gobble):

            # We deferred gobbling until now - fuse all points to be
            # gobbled and remove them from further consideration.

            # fuse dictionaries
            fused_to_gobble = collections.defaultdict(set)
            for (camn, frame_pt_idx, dd_idx) in all_to_gobble:
                fused_to_gobble[camn].add(dd_idx)

            # remove data to gobble
            for camn, dd_idx_set in fused_to_gobble.items():
                old_list = data_dict[camn]
                data_dict[camn] = [
                    item for (idx, item) in enumerate(old_list) if idx not in dd_idx_set
                ]

        # Take-back previous observations - starve this Kalman
        # object (which has higher error) so that 2 Kalman objects
        # don't start sharing all observations.
        _ = [
            self.live_tracked_objects[i].remove_previous_observation(debug1=debug2)
            for i in to_rewind
        ]

        # remove tracked objects from live list (when their error grows too large)
        kill_idxs = [
            idx for (idx, tro) in enumerate(self.live_tracked_objects) if tro.kill_me
        ]
        kill_idxs.sort()
        kill_idxs.reverse()
        newly_dead = [self.live_tracked_objects.pop(i) for i in kill_idxs]
        _ = [tro.kill() for tro in newly_dead]
        self.dead_tracked_objects.extend(newly_dead)
        self._flush_dead_queue()
        return data_dict

    def join_new_obj(
        self,
        frame,
        first_observation_orig_units,
        first_observation_Lcoords_orig_units,
        first_observation_camns,
        first_observation_idxs,
        debug=0,
    ):
        obj_id = self.cur_obj_id
        self.cur_obj_id += 1

        tro = TrackedObject(
            self.reconstructor,
            obj_id,
            frame,
            first_observation_orig_units,
            first_observation_Lcoords_orig_units,
            first_observation_camns,
            first_observation_idxs,
            kalman_model=self.kalman_model,
            save_all_data=self.save_all_data,
            area_threshold=self.area_threshold,
            area_threshold_for_orientation=self.area_threshold_for_orientation,
            disable_image_stat_gating=self.disable_image_stat_gating,
            orientation_consensus=self.orientation_consensus,
            fake_timestamp=self.fake_timestamp,
        )
        self.live_tracked_objects.append(tro)

    def kill_all_trackers(self):
        _ = [tro.kill() for tro in self.live_tracked_objects]
        while len(self.live_tracked_objects):
            self.dead_tracked_objects.append(self.live_tracked_objects.pop())
        self._flush_dead_queue()

    def set_killed_tracker_callback(self, callback):
        self.kill_tracker_callbacks.append(callback)

    def set_flushed_callback(self, callback):
        self.flushed_callbacks.append(callback)

    def clear_flushed_callbacks(self):
        self.flushed_callbacks = []

    def _flush_dead_queue(self):
        while len(self.dead_tracked_objects):
            tro = self.dead_tracked_objects.pop(0)
            for callback in self.kill_tracker_callbacks:
                callback(tro)
        for callback in self.flushed_callbacks:
            callback()



#emacs, this is -*-Python-*- mode
#cython: language_level=2

import numpy
import numpy.dual
inf = numpy.inf

cdef double cinf
cinf = inf


import flydra_core.common_variables
import flydra_core.undistort
import xml.etree.ElementTree as ET
import warnings
import os

from libc.math cimport sqrt

cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x)
    bint npy_isinf(double x)

cdef float MINIMUM_ECCENTRICITY
MINIMUM_ECCENTRICITY = flydra_core.common_variables.MINIMUM_ECCENTRICITY

STRICT_WATER = int(os.environ.get('STRICT_WATER_HYPOTHESIS_TEST','0'))

class NoAcceptablePointFound(Exception):
    pass

def make_ReconstructHelper_from_rad_file(filename):
    params = {}
    exec(open(filename).read(),params)
    if params['K12'] != 0:
        raise NotImplementedError('need to properly roundtrip alpha_c != 0')
    helper = ReconstructHelper(
        params['K11'], params['K22'], params['K13'], params['K23'],
        params['kc1'], params['kc2'], params['kc3'], params['kc4'])
    return helper

def ReconstructHelper_from_xml(elem):
    assert ET.iselement(elem)
    assert elem.tag == "non_linear_parameters"

    args = []
    for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c']:
        args.append( float(elem.find(name).text) )
    args = tuple(args)
    helper = ReconstructHelper(*args)
    return helper

def make_ReconstructHelper(*args,**kw):
    return ReconstructHelper(*args,**kw)

cdef class ReconstructHelper:
    cdef readonly double fc1, fc2, cc1, cc2
    cdef readonly double fc1p, fc2p, cc1p, cc2p
    cdef readonly double k1, k2, k3, p1, p2
    cdef readonly double alpha_c
    cdef readonly int simple

    def __init__(self, fc1, fc2, cc1, cc2, k1, k2, p1, p2, alpha_c=0,
                 fc1p=None, fc2p=None, cc1p=None, cc2p=None, k3=None,
                 ):
        """create instance of ReconstructHelper

        ReconstructHelper(fc1, fc2, cc1, cc2, k1, k2, p1, p2 )
        where:
        fc - focal length
        cc - camera center
        k - radial distortion parameters (non-linear)
        p - tangential distortion parameters (non-linear)
        alpha_c - skew between X and Y pixel axes
        """
        self.fc1 = self.fc1p = fc1
        self.fc2 = self.fc2p = fc2
        self.cc1 = self.cc1p = cc1
        self.cc2 = self.cc2p = cc2
        self.k1 = k1
        self.k2 = k2
        if k3 is None:
            k3 = 0.0
        self.k3 = k3
        self.p1 = p1
        self.p2 = p2
        self.alpha_c = alpha_c

        # "simple" means that the forward and back normalizations from
        # pixel coords to normalized coords are identical.

        self.simple=1
        if fc1p is not None and fc1p!=fc1:
            self.simple = 0
            self.fc1p = fc1p
        if fc2p is not None and fc2p!=fc2:
            self.simple = 0
            self.fc2p = fc2p
        if cc1p is not None and cc1p!=cc1:
            self.simple = 0
            self.cc1p = cc1p
        if cc2p is not None and cc2p!=cc2:
            self.simple = 0
            self.cc2p = cc2p
        if not self.simple and self.alpha_c != 0:
            raise NotImplementedError('skew not tested with non-simple distortion model')

    def __reduce__(self):
        """this allows ReconstructHelper to be pickled"""
        args = (self.fc1, self.fc2, self.cc1, self.cc2,
                self.k1, self.k2, self.p1, self.p2, self.alpha_c,
                self.fc1p, self.fc2p, self.cc1p, self.cc2p, self.k3)
        return (make_ReconstructHelper, args)

    def add_element(self, parent):
        """add self as XML element to parent"""
        assert ET.iselement(parent)
        elem = ET.SubElement(parent,"non_linear_parameters")

        if self.k3 != 0.0:
            raise NotImplementedError('k3 != 0.0')

        for name in ['fc1','fc2','cc1','cc2','k1','k2','p1','p2','alpha_c',
                     'fc1p', 'fc2p', 'cc1p', 'cc2p',
                     ]:
            e = ET.SubElement(elem, name)
            e.text = repr( getattr(self,name) )

    def as_obj_for_json(self):
        result = {}
        for name in ['fc1','fc2','cc1','cc2','k1','k2','k3','p1','p2','alpha_c',
                     'fc1p', 'fc2p', 'cc1p', 'cc2p']:
            result[name] = getattr(self,name)
        return result

    def save_to_rad_file( self, fname, comments=None ):
        if not self.simple:
            raise ValueError('cannot encode non-simple intrinsic model as a .rad file')
        rad_fd = open(fname,'w')
        K = self.get_K()
        nlparams = self.get_nlparams()
        k1, k2, p1, p2, k3 = nlparams
        rad_fd.write('K11 = %s;\n'%repr(K[0,0]))
        rad_fd.write('K12 = %s;\n'%repr(K[0,1]))
        rad_fd.write('K13 = %s;\n'%repr(K[0,2]))
        rad_fd.write('K21 = %s;\n'%repr(K[1,0]))
        rad_fd.write('K22 = %s;\n'%repr(K[1,1]))
        rad_fd.write('K23 = %s;\n'%repr(K[1,2]))
        rad_fd.write('K31 = %s;\n'%repr(K[2,0]))
        rad_fd.write('K32 = %s;\n'%repr(K[2,1]))
        rad_fd.write('K33 = %s;\n'%repr(K[2,2]))
        rad_fd.write('\n')
        rad_fd.write('kc1 = %s;\n'%repr(k1))
        rad_fd.write('kc2 = %s;\n'%repr(k2))
        rad_fd.write('kc3 = %s;\n'%repr(p1))
        rad_fd.write('kc4 = %s;\n'%repr(p2))
        rad_fd.write('kc5 = %s;\n'%repr(k3))
        rad_fd.write('\n')
        if comments is not None:
            comments = str(comments)
        rad_fd.write("comments = '%s';\n"%comments)
        rad_fd.write('\n')
        rad_fd.close()

    def __richcmp__(self,other,op):
        # cmp op
        #  < 0
        # <= 1
        # == 2
        # != 3
        #  > 4
        # >= 5

        if isinstance(other, ReconstructHelper):
            isequal = (numpy.allclose(self.get_K(),other.get_K()) and
                       numpy.allclose(self.get_Kp(),other.get_Kp()) and
                       numpy.allclose(self.get_nlparams(),other.get_nlparams()) )
        else:
            isequal = False

        if op in [1,2,5]:
            result = isequal
        elif op == 3:
            result = not isequal
        else:
            result = False
        return result

    def get_K(self):
        # See
        # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html
        K = numpy.array((( self.fc1, self.alpha_c*self.fc1, self.cc1),
                         ( 0,        self.fc2,              self.cc2),
                         ( 0,        0,                     1       )))
        return K

    def get_Kp(self):
        K = numpy.array((( self.fc1p, 0,                     self.cc1p),
                         ( 0,         self.fc2p,             self.cc2p),
                         ( 0,         0,                     1       )))
        return K

    def get_nlparams(self):
        return (self.k1, self.k2, self.p1, self.p2, self.k3)

    def undistort(self, double x_kk, double y_kk, int n_iterations=5):
        """undistort 2D coordinate pair

        Iteratively performs an undistortion using camera intrinsic
        parameters.

        Implementation translated from CalTechCal.

        See also the OpenCV reference manual, which has the equation
        used.
        """

        cdef double xl, yl

        cdef double xd, yd, x, y
        cdef double r_2, k_radial, delta_x, delta_y
        cdef int i

        # undoradial.m / CalTechCal/normalize.m

        xd = ( x_kk - self.cc1 ) / self.fc1
        yd = ( y_kk - self.cc2 ) / self.fc2

        xd = xd - self.alpha_c * yd

        # comp_distortion_oulu.m

        # initial guess
        x = xd
        y = yd

        for i from 0<=i<n_iterations:
            r_2 = x*x + y*y
            k_radial = 1.0 + r_2*(self.k1 + r_2*(self.k2 + r_2*self.k3))
            delta_x = 2.0 * (self.p1)*x*y + (self.p2)*(r_2 + 2.0*x*x)
            delta_y = (self.p1) * (r_2 + 2.0*y*y)+2.0*(self.p2)*x*y
            x = (xd-delta_x)/k_radial
            y = (yd-delta_y)/k_radial

        # undoradial.m

        xl = (self.fc1p)*x + (self.fc1p*self.alpha_c)*y + (self.cc1p)
        yl = (self.fc2p)*y + (self.cc2p)
        return (xl, yl)

    def distort(self, double xl, double yl):
        """distort 2D coordinate pair"""

        cdef double x, y, r_2, term1, xd, yd

        assert self.alpha_c==0 # there's no way the following code works otherwise

        x = ( xl - self.cc1p ) / self.fc1p
        y = ( yl - self.cc2p ) / self.fc2p

        r_2 = x*x + y*y
        r_4 = r_2**2
        r_6 = r_2*r_4
        term1 = self.k1*r_2 + self.k2*r_4 + self.k3*r_6

        # OpenCV manual (chaper 6, "3D reconstruction", "camera
        # calibration" section) seems to disagree with
        # http://www.vision.caltech.edu/bouguetj/calib_doc/htmls/parameters.html

        # Furthermore, implementations in cvundistort.cpp seem still
        # unrelated.  Finally, project_points2.m in Bouguet's code is
        # consistent with his webpage and this below.

        xd = x + x*term1 + (2*self.p1*x*y + self.p2*(r_2+2*x**2))
        yd = y + y*term1 + (self.p1*(r_2+2*y**2) + 2*self.p2*x*y)

        xd = (self.fc1)*xd + (self.cc1)
        yd = (self.fc2)*yd + (self.cc2)

        return (xd, yd)

    def undistort_image(self, numpy_image ):
        assert len(numpy_image.shape)==2
        assert numpy_image.dtype == numpy.uint8

        f = self.fc1, self.fc2 # focal length
        c = self.cc1, self.cc2 # center
        k = self.k1, self.k2, self.p1, self.p2, self.k3  # NL terms: r^2, r^4, tan1, tan2, r^6

        assert self.simple
        assert self.alpha_c==0

        imnew = flydra_core.undistort.rect(numpy_image, f=f, c=c, k=k) # perform the undistortion
        imnew = imnew.astype(numpy.uint8)
        return imnew

cdef int gave_water_warning
gave_water_warning = 0

def hypothesis_testing_algorithm__find_best_3d( object recon, object found_data_dict,
                                                double ACCEPTABLE_DISTANCE_PIXELS,
                                                int debug=0, Py_ssize_t max_n_cams=5,
                                                int with_water = 0):
    """Use hypothesis testing algorithm to find best 3D point

    Finds combination of cameras which uses the most number of cameras
    while minimizing mean reprojection error. Algorithm used accepts
    any camera combination with reprojection error less than the
    variable ACCEPTABLE_DISTANCE_PIXELS.

    """
    cdef int n_cams, best_n_cams, alloc_n_cams
    cdef int i
    cdef int missing_cam_data
    cdef double alpha
    cdef double x, y, orig_x, orig_y, new_x, new_y
    cdef double dist, mean_dist, least_err

    #cdef double *least_err_by_n_cameras # fake dict (index = key)

    svd = numpy.dual.svd # eliminate global name lookup

    cam_ids = recon.cam_ids # shorthand
    max_n_cams = min(len(cam_ids), max_n_cams)

    # Initialize least_err_by_n_cameras to be infinity.  Note that
    # values at 0th and 1st index will always remain infinity.

    least_err_by_n_cameras = [cinf]*(max_n_cams+1) # allow 0-based indexing to last camera
    allA = numpy.zeros( (2*(len(cam_ids)+1),4), dtype=numpy.float64)
    bad_cam_ids = []
    cam_id2idx = {}
    all2d = {}
    Pmat_fastnx = recon.Pmat # shorthand

    global gave_water_warning
    if with_water:
        if STRICT_WATER:
            raise NotImplementedError('water and hypothesis testing not yet implemented')
        if not gave_water_warning:
            warnings.warn('_reconstruct_utils: Hypothesis test intersection done '
                          'without refraction correction. Result will be wrong. '
                          'Set environment variable STRICT_WATER_HYPOTHESIS_TEST '
                          'to raise an error rather than give this warning.')
            gave_water_warning = 1
    for i,cam_id in enumerate(cam_ids):
        cam_id2idx[cam_id] = i

        # do we have incoming data?
        try:
            value_tuple = found_data_dict[cam_id]
        except KeyError:
            bad_cam_ids.append( cam_id )
            continue # don't build this row

        # was a 2d point found?
        xy_values = value_tuple[:2]
        x,y = xy_values
        if npy_isnan(x):
            bad_cam_ids.append( cam_id )
            continue # don't build this row

        # Similar to code in reconstruct.Reconstructor.find3d()

        Pmat = Pmat_fastnx[cam_id] # Pmat is 3 rows x 4 columns
        row3 = Pmat[2,:]
        allA[ i*2, : ] = x*row3 - Pmat[0,:]
        allA[ i*2+1, :]= y*row3 - Pmat[1,:]

        all2d[cam_id] = xy_values

    cam_ids_for_least_err = {}
    X_for_least_err = {}
    for n_cams from 2<=n_cams<=max_n_cams:

        # Calculate in least reprojection error starting with all
        # possible combinations of 2 cameras, then start increasing
        # the number of cameras used.  For each number of cameras,
        # determine if there exists a combination with an acceptable
        # reprojection error.

        alpha = 1.0/n_cams

        # Can we short-circuit the rest of these computations?
        if not npy_isinf(least_err_by_n_cameras[n_cams-1]): # if it's infinity, it must be n_cams 0 or 1.
            # If we've calculated error for 2 less than n_cams
            if least_err_by_n_cameras[n_cams-1] > ACCEPTABLE_DISTANCE_PIXELS:
                if debug>5:
                    print 'HYPOTHESIS TEST -    shorting for n_cams %d (this=%f, ACCEPTABLE_DISTANCE_PIXELS=%f)'%(
                        n_cams,least_err_by_n_cameras[n_cams-1],ACCEPTABLE_DISTANCE_PIXELS)
                # and if the error for 1 less is too large, don't bother with more.
                break

        for cam_ids_used in recon.cam_combinations_by_size[n_cams]:
            missing_cam_data = 0 #False
            good_A_idx = []
            if debug>5:
                assert len(cam_ids_used)==n_cams
            for cam_id in cam_ids_used:
                if cam_id in bad_cam_ids:
                    missing_cam_data = 1 #True
                    break
                else:
                    i = cam_id2idx[cam_id]
                    good_A_idx.extend( (i*2, i*2+1) )
            if missing_cam_data == 1:
                continue

            # find 3D point
            A = allA[good_A_idx,:]
            u,d,vt=svd(A)
            X = vt[-1,:]/vt[-1,3] # normalize

            # calculate reprojection error
            mean_dist = 0.0
            for cam_id in cam_ids_used:
                orig_x,orig_y = all2d[cam_id]
                Pmat = Pmat_fastnx[cam_id]
                new_xyw = numpy.dot( Pmat, X ) # reproject 3d to 2d
                new_x, new_y = new_xyw[0:2]/new_xyw[2]

                dist = sqrt((orig_x-new_x)**2 + (orig_y-new_y)**2)
                mean_dist = mean_dist + dist*alpha
            if debug>5:
                print 'HYPOTHESIS TEST - mean_dist = %f for cam_ids_used = %s (always pt 0)'%(mean_dist,str(cam_ids_used))

            least_err = least_err_by_n_cameras[n_cams]
            if mean_dist < least_err:
                least_err_by_n_cameras[n_cams] = mean_dist
                cam_ids_for_least_err[n_cams] = cam_ids_used
                X_for_least_err[n_cams] = X[:3]

    # now we have the best estimate for 2 views, 3 views, ...
    best_n_cams = 2
    least_err = least_err_by_n_cameras[2]
    if debug>5:
        print 'HYPOTHESIS TEST - least_err, ACCEPTABLE_DISTANCE_PIXELS: %f, %f'%(least_err,ACCEPTABLE_DISTANCE_PIXELS)
    mean_dist = least_err
    if not (least_err < ACCEPTABLE_DISTANCE_PIXELS):
        raise NoAcceptablePointFound('least error was %f'%least_err)

    for n_cams from 3 <= n_cams <= max_n_cams:
        least_err = least_err_by_n_cameras[n_cams]
        if debug>5:
            print 'HYPOTHESIS TEST - n_cams %d: %f'%(n_cams,least_err)
        if npy_isinf(least_err):
            break # if we don't have e.g. 4 cameras, we won't have 5
        if least_err < ACCEPTABLE_DISTANCE_PIXELS:
            mean_dist = least_err
            best_n_cams = n_cams

    # now calculate final values
    cam_ids_used = cam_ids_for_least_err[best_n_cams]
    if with_water:
        # Even though (for speed reasons) we did not use proper
        # refraction-correct code above, we now recompute X using
        # refraction.
        cam_ids_and_points2d = []
        for cam_id in cam_ids_used:
            xy = found_data_dict[cam_id][:2]
            cam_ids_and_points2d.append(( cam_id, xy ))

        X = recon.find3d(cam_ids_and_points2d,
                         undistort=False, # points are already undistorted
                         return_line_coords = False,
                         )
    else:
        X = X_for_least_err[best_n_cams]
    del X_for_least_err

    # calculate line3d
    P = []
    for cam_id in cam_ids_used:
        x,y,area,slope,eccentricity, p1,p2,p3,p4 = found_data_dict[cam_id]
        if eccentricity > MINIMUM_ECCENTRICITY and not numpy.isnan(p1):
                P.append( (p1,p2,p3,p4) )
    if len(P) < 2:
        Lcoords = None
    else:
        P = numpy.array(P)
        try:
            u,d,vt=svd(P,full_matrices=True)
        except numpy.linalg.LinAlgError, err:
            print 'SVD error, P=',repr(P)
            Lcoords = None
        except:
            print 'Error on P'
            print P
            raise
        else:
            P = vt[0,:] # P,Q are planes (take row because this is transpose(V))
            Q = vt[1,:]

            # directly to Pluecker line coordinates
            Lcoords = ( -(P[3]*Q[2]) + P[2]*Q[3],
                          P[3]*Q[1]  - P[1]*Q[3],
                        -(P[2]*Q[1]) + P[1]*Q[2],
                        -(P[3]*Q[0]) + P[0]*Q[3],
                        -(P[2]*Q[0]) + P[0]*Q[2],
                        -(P[1]*Q[0]) + P[0]*Q[1] )
            if npy_isnan(Lcoords[0]):
                Lcoords = None
    ## c_lib.free(least_err_by_n_cameras)
    return X, Lcoords, cam_ids_used, mean_dist

## def undistort_image( char* src, int srcstep,
##                      char* dst, int dststep,
##                      int width, int height,
##                      float u0, float v0,
##                      float fx, float fy,
##                      float k1, float k2,
##                      float p1, float p2,
##                      int cn):
##     """derived from cvundistort.cpp"""

##     cdef int u, v, i
##     cdef float y, y2, ky, k2y, _2p1y, _3p1y2, p2y2, _fy, _fx
##     cdef float x, x2, kx, d, _u, _v, iu, iv, ifx, ify
##     cdef float a0, a1, b0, b1
##     cdef char* ptr
##     float t0, t1

##     _fx = 1.f/fx
##     _fy = 1.f/fy

##     for v from 0 <= v < height:
##         dst = dst + dststep;

##         y = (v - v0)*_fy
##         y2 = y*y
##         ky = 1 + (k1 + k2*y2)*y2;
##         k2y = 2*k2*y2
##         _2p1y = 2*p1*y
##         _3p1y2 = 3*p1*y2
##         p2y2 = p2*y2

##         for u from 0 <= u < width:
##             x = (u - u0)*_fx
##             x2 = x*x
##             kx = (k1 + k2*x2)*x2
##             d = kx + ky + k2y*x2
##             _u = fx*(x*(d + _2p1y) + p2y2 + (3*p2)*x2) + u0
##             _v = fy*(y*(d + (2*p2)*x) + _3p1y2 + p1*x2) + v0
##             iu = cvRound(_u*(1 << ICV_WARP_SHIFT))
##             iv = cvRound(_v*(1 << ICV_WARP_SHIFT))
##             ifx = iu & ICV_WARP_MASK
##             ify = iv & ICV_WARP_MASK

##             a0 = icvLinearCoeffs[ifx*2]
##             a1 = icvLinearCoeffs[ifx*2 + 1]
##             b0 = icvLinearCoeffs[ify*2]
##             b1 = icvLinearCoeffs[ify*2 + 1]

##             if( <unsigned>iv < <unsigned>(height - 1) &&
##                 <unsigned>iu < <unsigned>(width - 1) ):
##                 ptr = src + iv*srcstep + iu*cn
##                 for i from 0 <= i < cn:
##                     t0 = a1*CV_8TO32F(ptr[i]) + a0*CV_8TO32F(ptr[i+cn])
##                     t1 = a1*CV_8TO32F(ptr[i+srcstep]) + a0*CV_8TO32F(ptr[i + srcstep + cn])
##                     dst[u*cn + i] = <char>cvRound(b1*t0 + b0*t1)
##             else:
##                 for i from 0 <= i < cn:
##                     dst[u*cn + i] = 0

find_best_3d = hypothesis_testing_algorithm__find_best_3d # old name



import tables as PT

# This describes indexes in the "pt_undistorted" tuple.  These are
# used in MainBrain.py, flydra_tracker.py, and kalmanize.py
PT_TUPLE_IDX_X = 0
PT_TUPLE_IDX_Y = 1
PT_TUPLE_IDX_AREA = 2
PT_TUPLE_IDX_SLOPE = 3
PT_TUPLE_IDX_ECCENTRICITY = 4
# 3D coordinates of plane formed by camera center and slope line
# centered on object.
PT_TUPLE_IDX_P1 = 5
PT_TUPLE_IDX_P2 = 6
PT_TUPLE_IDX_P3 = 7
PT_TUPLE_IDX_P4 = 8
PT_TUPLE_IDX_LINE_FOUND = 9
PT_TUPLE_IDX_FRAME_PT_IDX = 10
PT_TUPLE_IDX_CUR_VAL_IDX = 11
PT_TUPLE_IDX_MEAN_VAL_IDX = 12
PT_TUPLE_IDX_SUMSQF_VAL_IDX = 13

WIRE_ORDER_CUR_VAL_IDX = 6
WIRE_ORDER_MEAN_VAL_IDX = 7
WIRE_ORDER_SUMSQF_VAL_IDX = 8

# 2D data format for PyTables:
class Info2D(PT.IsDescription):
    camn = PT.UInt16Col(pos=0)
    frame = PT.Int64Col(pos=1)
    timestamp = PT.FloatCol(
        pos=2
    )  # when the image trigger happened (returned by timestamp modeler on MainBrain)
    cam_received_timestamp = PT.FloatCol(
        pos=3
    )  # when the image was acquired by flydra software (on camera computer)
    x = PT.Float32Col(pos=4)
    y = PT.Float32Col(pos=5)
    area = PT.Float32Col(pos=6)
    slope = PT.Float32Col(pos=7)
    eccentricity = PT.Float32Col(pos=8)
    frame_pt_idx = PT.UInt8Col(
        pos=9
    )  # index of point if there were > 1 points in frame
    cur_val = PT.UInt8Col(pos=10)
    mean_val = PT.Float32Col(pos=11)
    sumsqf_val = PT.Float32Col(pos=12)  # estimate of <x^2> (running_sumsqf)


class TextLogDescription(PT.IsDescription):
    mainbrain_timestamp = PT.FloatCol(pos=0)
    cam_id = PT.StringCol(255, pos=1)
    host_timestamp = PT.FloatCol(pos=2)
    message = PT.StringCol(255, pos=3)


class CamSyncInfo(PT.IsDescription):
    cam_id = PT.StringCol(256, pos=0)
    camn = PT.UInt16Col(pos=1)
    hostname = PT.StringCol(2048, pos=2)


class HostClockInfo(PT.IsDescription):
    remote_hostname = PT.StringCol(255, pos=0)
    start_timestamp = PT.FloatCol(pos=1)
    remote_timestamp = PT.FloatCol(pos=2)
    stop_timestamp = PT.FloatCol(pos=3)


class TriggerClockInfo(PT.IsDescription):
    start_timestamp = PT.FloatCol(pos=0)
    framecount = PT.Int64Col(pos=1)
    tcnt = PT.UInt16Col(pos=2)
    stop_timestamp = PT.FloatCol(pos=3)


class MovieInfo(PT.IsDescription):
    cam_id = PT.StringCol(16, pos=0)
    filename = PT.StringCol(255, pos=1)
    approx_start_frame = PT.Int64Col(pos=2)
    approx_stop_frame = PT.Int64Col(pos=3)


class ExperimentInfo(PT.IsDescription):
    uuid = PT.StringCol(32, pos=0)
